# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2022 Scipp contributors (https://github.com/scipp)
import datetime
import pytest
import scipp as sc
import platform
import warnings
from typing import List, Tuple, Optional, Union
import numpy as np
from .nexus_helpers import NexusBuilder, Stream, Log, EventData
import multiprocessing as mp
from scippneutron.data_streaming._consumer_type import ConsumerType
from scippneutron.data_streaming._warnings import (UnknownFlatbufferIdWarning,
                                                   BufferSizeWarning)
from cmath import isclose

if platform.system() == "Darwin":
    pytest.skip(
        "Streaming tests disabled on MacOS because they tend to hang"
        " indefinitely for unknown reasons",
        allow_module_level=True)

try:
    import streaming_data_types  # noqa: F401
    from confluent_kafka import TopicPartition, KafkaError  # noqa: F401
    from scippneutron.data_streaming.data_stream import _data_stream  # noqa: E402
    from streaming_data_types.eventdata_ev42 import serialise_ev42  # noqa: E402
    from streaming_data_types.run_start_pl72 import serialise_pl72
    from streaming_data_types.logdata_f142 import serialise_f142
    from streaming_data_types.timestamps_tdct import serialise_tdct
    from streaming_data_types.sample_environment_senv import serialise_senv
    from streaming_data_types.sample_environment_senv import Location
    from scippneutron.data_streaming._consumer import RunStartError
    from scippneutron.data_streaming.data_stream import StopTime
except ImportError:
    pytest.skip("Kafka or Serialisation module is unavailable", allow_module_level=True)


class FakeKafkaError:

    def __init__(self, kafka_error_code: int):
        self._error_code = kafka_error_code

    def code(self):
        return self._error_code

    def __str__(self):
        return f"FakeKafkaError: code {self._error_code}"


class FakeMessage:

    def __init__(self,
                 payload: bytes,
                 error_code: Optional[int] = None,
                 timestamp: int = 0):
        self._payload = payload
        self._error_code = error_code
        self._timestamp = timestamp

    def value(self):
        return self._payload

    def error(self) -> Optional[KafkaError]:
        if self._error_code is not None:
            return FakeKafkaError(self._error_code)
        return None

    def timestamp(self) -> Tuple[None, int]:
        return None, self._timestamp


class FakeQueryConsumer:

    def __init__(self,
                 instrument_name: str = "",
                 low_and_high_offset: Tuple[int, int] = (2, 10),
                 streams: List[Stream] = None,
                 start_time: Optional[Union[int, datetime.datetime]] = None,
                 stop_time: Optional[Union[int, datetime.datetime]] = None,
                 nexus_structure: Optional[str] = None):
        self._instrument_name = instrument_name
        self._low_and_high_offset = low_and_high_offset
        self._streams = streams
        self.queried_topics = []
        self.queried_timestamp = None
        self._start_time = start_time
        self._stop_time = stop_time
        if self._start_time is None:
            self._start_time = datetime.datetime.now()
        if nexus_structure is None:
            builder = NexusBuilder()
            builder.add_instrument(self._instrument_name)
            if self._streams is not None:
                for stream in self._streams:
                    builder.add_stream(stream)
            self._nexus_structure = builder.json_string
        else:
            self._nexus_structure = nexus_structure

    @staticmethod
    def assign(partitions: List[TopicPartition]):
        pass

    def get_watermark_offsets(self, partition: TopicPartition) -> Tuple[int, int]:
        return self._low_and_high_offset

    def get_topic_partitions(self,
                             topic: str,
                             offset: int = -1) -> List[TopicPartition]:
        self.queried_topics.append(topic)
        return [TopicPartition(topic, partition=0, offset=offset)]

    def poll(self, timeout=2.) -> FakeMessage:
        return FakeMessage(
            serialise_pl72("",
                           "",
                           start_time=self._start_time,
                           stop_time=self._stop_time,
                           nexus_structure=self._nexus_structure))

    def seek(self, partition: TopicPartition):
        pass

    def offsets_for_times(self, partitions: List[TopicPartition]):
        self.queried_timestamp = partitions[0].offset
        return partitions


# Short time to use for buffer emit and data_stream interval in tests
# pass or fail fast!
SHORT_TEST_INTERVAL = 100. * sc.Unit('milliseconds')
# Small buffer of 20 events is sufficient for the tests
TEST_BUFFER_SIZE = 20

TEST_STREAM_ARGS = {
    "kafka_broker": "broker",
    "topics": ["topic"],
    "interval": SHORT_TEST_INTERVAL,
    "event_buffer_size": TEST_BUFFER_SIZE,
    "slow_metadata_buffer_size": TEST_BUFFER_SIZE,
    "fast_metadata_buffer_size": TEST_BUFFER_SIZE,
    "chopper_buffer_size": TEST_BUFFER_SIZE,
    "consumer_type": ConsumerType.FAKE,
    "timeout": 10. * sc.units.s
}
# "timeout" arg: if something gets broken then this makes sure the
# test will not get stuck in the _data_stream loop indefinitely.
# A TimeoutError is raised by _data_stream if the timeout occurs.


@pytest.fixture(scope="function")
def queues():
    # Specify to start the process using the "spawn" method, otherwise
    # on Linux the default is to fork the Python interpreter which
    # is "problematic" in a multithreaded process, in our case the use of
    # asyncio means the process is multithreaded.
    # See documentation:
    # https://docs.python.org/3/library/multiprocessing.html#contexts-and-start-methods
    ctx = mp.get_context("spawn")
    return ctx.Queue(), ctx.Queue(), ctx.Queue()


@pytest.mark.asyncio
async def test_data_stream_returns_data_from_single_event_message(queues):
    data_queue, worker_instruction_queue, test_message_queue = queues
    time_of_flight = np.array([1., 2., 3.])
    detector_ids = np.array([4, 5, 6])
    test_message = FakeMessage(
        serialise_ev42("detector", 0, 0, time_of_flight, detector_ids))
    test_message_queue.put(test_message)

    reached_assert = False
    async for data in _data_stream(data_queue,
                                   worker_instruction_queue,
                                   halt_after_n_data_chunks=1,
                                   test_message_queue=test_message_queue,
                                   query_consumer=FakeQueryConsumer(),
                                   **TEST_STREAM_ARGS):
        assert np.allclose(data.coords['tof'].values, time_of_flight)
        reached_assert = True
    assert reached_assert


@pytest.mark.asyncio
async def test_data_stream_returns_data_from_multiple_event_messages(queues):
    data_queue, worker_instruction_queue, test_message_queue = queues
    first_tof = np.array([1., 2., 3.])
    first_detector_ids = np.array([4, 5, 6])
    first_test_message = serialise_ev42("detector", 0, 0, first_tof, first_detector_ids)
    second_tof = np.array([1., 2., 3.])
    second_detector_ids = np.array([4, 5, 6])
    second_test_message = serialise_ev42("detector", 0, 0, second_tof,
                                         second_detector_ids)
    test_message_queue.put(FakeMessage(first_test_message))
    test_message_queue.put(FakeMessage(second_test_message))

    reached_asserts = False
    async for data in _data_stream(data_queue,
                                   worker_instruction_queue,
                                   halt_after_n_data_chunks=1,
                                   test_message_queue=test_message_queue,
                                   query_consumer=FakeQueryConsumer(),
                                   **TEST_STREAM_ARGS):
        expected_tofs = np.concatenate((first_tof, second_tof))
        assert np.allclose(data.coords['tof'].values, expected_tofs)
        expected_ids = np.concatenate((first_detector_ids, second_detector_ids))
        assert np.array_equal(data.coords['detector_id'].values, expected_ids)
        reached_asserts = True
    assert reached_asserts


@pytest.mark.asyncio
async def test_warn_if_unrecognised_message_was_encountered(queues):
    warnings.filterwarnings("error")
    data_queue, worker_instruction_queue, test_message_queue = queues
    # First 4 bytes of the message payload are the FlatBuffer schema identifier
    # "abcd" does not correspond to a FlatBuffer schema for data
    # that scipp is interested in
    test_message = b"abcd0000"

    with pytest.warns(UnknownFlatbufferIdWarning):
        test_message_queue.put(FakeMessage(test_message))
        async for _ in _data_stream(data_queue,
                                    worker_instruction_queue,
                                    halt_after_n_warnings=1,
                                    test_message_queue=test_message_queue,
                                    query_consumer=FakeQueryConsumer(),
                                    **TEST_STREAM_ARGS):
            test_message_queue.put(FakeMessage(test_message))


@pytest.mark.asyncio
async def test_warn_on_buffer_size_exceeded_by_single_message(queues):
    data_queue, worker_instruction_queue, test_message_queue = queues
    buffer_size_2_events = 2
    time_of_flight = np.array([1., 2., 3.])
    detector_ids = np.array([4, 5, 6])
    test_message = serialise_ev42("detector", 0, 0, time_of_flight, detector_ids)

    test_steam_args = TEST_STREAM_ARGS.copy()
    test_steam_args["event_buffer_size"] = buffer_size_2_events

    with pytest.warns(BufferSizeWarning):
        test_message_queue.put(FakeMessage(test_message))
        async for _ in _data_stream(data_queue,
                                    worker_instruction_queue,
                                    halt_after_n_warnings=1,
                                    test_message_queue=test_message_queue,
                                    query_consumer=FakeQueryConsumer(),
                                    **test_steam_args):
            test_message_queue.put(FakeMessage(test_message))


@pytest.mark.asyncio
async def test_data_returned_when_buffer_size_exceeded_by_event_messages(queues):
    # Messages cumulatively exceed the buffer size, data_stream
    # will return multiple chunks of data to clear the buffer
    # between messages.
    data_queue, worker_instruction_queue, test_message_queue = queues
    first_tof = np.array([1., 2., 3.])
    first_detector_ids = np.array([4, 5, 6])
    first_test_message = serialise_ev42("detector", 0, 0, first_tof, first_detector_ids)
    second_tof = np.array([7., 8., 9.])
    second_detector_ids = np.array([4, 5, 6])
    second_test_message = serialise_ev42("detector", 0, 0, second_tof,
                                         second_detector_ids)

    # Event data buffer size is 5, so the second message
    # will not fit in the buffer
    test_stream_args = TEST_STREAM_ARGS.copy()
    test_stream_args["event_buffer_size"] = 5
    test_stream_args["run_info_topic"] = "run_info_topic"

    reached_asserts = False
    n_chunks = 0
    async for data in _data_stream(data_queue,
                                   worker_instruction_queue,
                                   halt_after_n_data_chunks=3,
                                   test_message_queue=test_message_queue,
                                   query_consumer=FakeQueryConsumer(),
                                   **test_stream_args):
        # n_chunks == 0 zeroth chunk contains data
        # from run start message
        if n_chunks == 0:
            test_message_queue.put(FakeMessage(first_test_message))
            test_message_queue.put(FakeMessage(second_test_message))
        elif n_chunks == 1:
            # Contain event data from first message
            assert np.allclose(data.coords['tof'].values, first_tof)
        elif n_chunks == 2:
            # Contain event data from second message
            assert np.allclose(data.coords['tof'].values, second_tof)
            reached_asserts = True

        n_chunks += 1
    assert reached_asserts


@pytest.mark.asyncio
async def test_data_are_loaded_from_run_start_message(queues):
    data_queue, worker_instruction_queue, test_message_queue = queues
    run_info_topic = "fake_topic"
    reached_assert = False
    test_instrument_name = "DATA_STREAM_TEST"
    async for data in _data_stream(
            data_queue,
            worker_instruction_queue,
            run_info_topic=run_info_topic,
            halt_after_n_data_chunks=0,
            test_message_queue=test_message_queue,
            query_consumer=FakeQueryConsumer(test_instrument_name),
            **TEST_STREAM_ARGS):
        assert data["instrument_name"].value == test_instrument_name
        reached_assert = True
    assert reached_assert


@pytest.mark.asyncio
async def test_error_raised_if_no_run_start_message_available(queues):
    data_queue, worker_instruction_queue, test_message_queue = queues
    run_info_topic = "fake_topic"
    test_instrument_name = "DATA_STREAM_TEST"

    # Low and high offset are the same value, indicates there are
    # no messages available in the partition
    low_and_high_offset = (0, 0)
    with pytest.raises(RunStartError):
        async for _ in _data_stream(data_queue,
                                    worker_instruction_queue,
                                    run_info_topic=run_info_topic,
                                    halt_after_n_data_chunks=0,
                                    test_message_queue=test_message_queue,
                                    query_consumer=FakeQueryConsumer(
                                        test_instrument_name, low_and_high_offset),
                                    **TEST_STREAM_ARGS):
            pass


@pytest.mark.asyncio
async def test_error_if_both_topics_and_run_start_topic_not_specified(queues):
    data_queue, worker_instruction_queue, test_message_queue = queues

    test_stream_args = TEST_STREAM_ARGS.copy()
    test_stream_args["topics"] = None
    # At least one of "topics" and "run_start_topic" must be specified
    with pytest.raises(ValueError):
        async for _ in _data_stream(data_queue,
                                    worker_instruction_queue,
                                    run_info_topic=None,
                                    halt_after_n_data_chunks=0,
                                    **test_stream_args,
                                    query_consumer=FakeQueryConsumer(),
                                    test_message_queue=test_message_queue):
            pass


@pytest.mark.asyncio
async def test_specified_topics_override_run_start_message_topics(queues):
    # If "topics" argument is specified then they should be used, even if
    # a run start topic is provided
    data_queue, worker_instruction_queue, test_message_queue = queues
    test_topics = ["whiting", "snail", "porpoise"]
    topic_in_run_start_message = "test_topic"
    test_streams = [Stream("/entry", topic_in_run_start_message)]
    query_consumer = FakeQueryConsumer(streams=test_streams)
    test_stream_args = TEST_STREAM_ARGS.copy()
    test_stream_args["topics"] = test_topics
    async for _ in _data_stream(data_queue,
                                worker_instruction_queue,
                                run_info_topic=None,
                                query_consumer=query_consumer,
                                halt_after_n_data_chunks=0,
                                **test_stream_args,
                                test_message_queue=test_message_queue):
        pass
    assert not query_consumer.queried_topics, "Expected specified topics" \
                                              " to be used and none queried"


@pytest.mark.asyncio
async def test_data_stream_returns_metadata(queues):
    data_queue, worker_instruction_queue, test_message_queue = queues
    run_info_topic = "fake_topic"
    test_instrument_name = "DATA_STREAM_TEST"

    # The Kafka topics to get metadata from are recorded as "stream" objects in
    # the nexus_structure field of the run start message
    # There are currently 3 schemas for metadata, they have flatbuffer ids
    # f142, senv and tdct
    f142_source_name = "f142_source"
    f142_log_name = "f142_log"
    senv_source_name = "senv_source"
    senv_log_name = "senv_log"
    tdct_source_name = "tdct_source"
    tdct_log_name = "tdct_log"
    streams = [
        Stream(f"/entry/{f142_log_name}", "f142_topic", f142_source_name, "f142",
               "double", "m"),
        Stream(f"/entry/{senv_log_name}", "senv_topic", senv_source_name, "senv",
               "double", "m"),
        Stream(f"/entry/{tdct_log_name}", "tdct_topic", tdct_source_name, "tdct")
    ]

    test_stream_args = TEST_STREAM_ARGS.copy()
    test_stream_args["topics"] = None
    n_chunks = 0
    async for data in _data_stream(data_queue,
                                   worker_instruction_queue,
                                   run_info_topic=run_info_topic,
                                   query_consumer=FakeQueryConsumer(
                                       test_instrument_name, streams=streams),
                                   halt_after_n_data_chunks=2,
                                   **test_stream_args,
                                   test_message_queue=test_message_queue):
        data_from_stream = data

        if n_chunks == 0:
            # Fake receiving a Kafka message for each metadata schema
            # Do this after the run start message has been parsed, so that
            # a metadata buffer will have been created for each data source
            # described in the start message.
            f142_value = 26.1236
            f142_timestamp = 123456  # ns after epoch
            f142_test_message = serialise_f142(f142_value, f142_source_name,
                                               f142_timestamp)
            test_message_queue.put(FakeMessage(f142_test_message))
            senv_values = np.array([26, 127, 52])
            senv_timestamp_ns = 123000  # ns after epoch
            senv_timestamp = datetime.datetime.fromtimestamp(senv_timestamp_ns * 1e-9,
                                                             datetime.timezone.utc)
            senv_time_between_samples = 100  # ns
            senv_test_message = serialise_senv(senv_source_name, -1, senv_timestamp,
                                               senv_time_between_samples, 0,
                                               senv_values, Location.Start)
            test_message_queue.put(FakeMessage(senv_test_message))
            tdct_timestamps = np.array([1234, 2345, 3456])  # ns
            tdct_test_message = serialise_tdct(tdct_source_name, tdct_timestamps)
            test_message_queue.put(FakeMessage(tdct_test_message))

        n_chunks += 1

    assert isclose(data_from_stream.attrs[f142_source_name].value.values[0], f142_value)
    assert data_from_stream.attrs[f142_source_name].value.coords['time'].values[
        0] == np.array(f142_timestamp, dtype=np.dtype('datetime64[ns]'))
    assert np.array_equal(data_from_stream.attrs[senv_source_name].value.values,
                          senv_values)
    senv_expected_timestamps = np.array([
        senv_timestamp_ns, senv_timestamp_ns + senv_time_between_samples,
        senv_timestamp_ns + (2 * senv_time_between_samples)
    ],
                                        dtype=np.dtype('datetime64[ns]'))
    assert np.array_equal(
        data_from_stream.attrs[senv_source_name].value.coords['time'].values,
        senv_expected_timestamps)
    assert np.array_equal(data_from_stream.attrs[tdct_source_name].value.values,
                          tdct_timestamps)


@pytest.mark.asyncio
async def test_data_stream_returns_data_from_multiple_slow_metadata_messages(queues):
    data_queue, worker_instruction_queue, test_message_queue = queues
    run_info_topic = "fake_topic"
    test_instrument_name = "DATA_STREAM_TEST"

    # The Kafka topics to get metadata from are recorded as "stream" objects in
    # the nexus_structure field of the run start message
    f142_source_name = "f142_source"
    f142_log_name = "f142_log"
    streams = [
        Stream(f"/entry/{f142_log_name}", "f142_topic", f142_source_name, "f142",
               "double", "m"),
    ]

    test_stream_args = TEST_STREAM_ARGS.copy()
    test_stream_args["topics"] = None
    n_chunks = 0
    async for data in _data_stream(data_queue,
                                   worker_instruction_queue,
                                   run_info_topic=run_info_topic,
                                   query_consumer=FakeQueryConsumer(
                                       test_instrument_name, streams=streams),
                                   halt_after_n_data_chunks=2,
                                   **test_stream_args,
                                   test_message_queue=test_message_queue):
        data_from_stream = data

        if n_chunks == 0:
            # Fake receiving a Kafka message for each metadata schema
            # Do this after the run start message has been parsed, so that
            # a metadata buffer will have been created for each data source
            # described in the start message.
            f142_value_1 = 26.1236
            f142_timestamp_1 = 123456  # ns after epoch
            f142_test_message = serialise_f142(f142_value_1, f142_source_name,
                                               f142_timestamp_1)
            test_message_queue.put(FakeMessage(f142_test_message))
            f142_value_2 = 2.725
            f142_timestamp_2 = 234567  # ns after epoch
            f142_test_message = serialise_f142(f142_value_2, f142_source_name,
                                               f142_timestamp_2)
            test_message_queue.put(FakeMessage(f142_test_message))

        n_chunks += 1

    assert np.allclose(data_from_stream.attrs[f142_source_name].value.values,
                       np.array([f142_value_1, f142_value_2]))
    assert np.array_equal(
        data_from_stream.attrs[f142_source_name].value.coords['time'].values,
        np.array([f142_timestamp_1, f142_timestamp_2],
                 dtype=np.dtype('datetime64[ns]')))


@pytest.mark.asyncio
async def test_data_stream_returns_data_from_multiple_fast_metadata_messages(queues):
    data_queue, worker_instruction_queue, test_message_queue = queues
    run_info_topic = "fake_topic"
    test_instrument_name = "DATA_STREAM_TEST"

    # The Kafka topics to get metadata from are recorded as "stream" objects in
    # the nexus_structure field of the run start message
    senv_source_name = "senv_source"
    senv_log_name = "senv_log"
    streams = [
        Stream(f"/entry/{senv_log_name}", "senv_topic", senv_source_name, "senv",
               "double", "m"),
    ]

    test_stream_args = TEST_STREAM_ARGS.copy()
    test_stream_args["topics"] = None
    n_chunks = 0
    async for data in _data_stream(data_queue,
                                   worker_instruction_queue,
                                   run_info_topic=run_info_topic,
                                   query_consumer=FakeQueryConsumer(
                                       test_instrument_name, streams=streams),
                                   halt_after_n_data_chunks=2,
                                   **test_stream_args,
                                   test_message_queue=test_message_queue):
        data_from_stream = data

        if n_chunks == 0:
            # Fake receiving a Kafka message for each metadata schema
            # Do this after the run start message has been parsed, so that
            # a metadata buffer will have been created for each data source
            # described in the start message.
            senv_values_1 = np.array([26, 127, 52])
            senv_timestamp_ns_1 = 123000  # ns after epoch
            senv_timestamp = datetime.datetime.fromtimestamp(senv_timestamp_ns_1 * 1e-9,
                                                             datetime.timezone.utc)
            senv_time_between_samples = 100  # ns
            senv_test_message = serialise_senv(senv_source_name, -1, senv_timestamp,
                                               senv_time_between_samples, 0,
                                               senv_values_1, Location.Start)
            test_message_queue.put(FakeMessage(senv_test_message))
            senv_values_2 = np.array([3832, 324, 3])
            senv_timestamp_ns_2 = 234000  # ns after epoch
            senv_timestamp = datetime.datetime.fromtimestamp(senv_timestamp_ns_2 * 1e-9,
                                                             datetime.timezone.utc)
            senv_test_message = serialise_senv(senv_source_name, -1, senv_timestamp,
                                               senv_time_between_samples, 0,
                                               senv_values_2, Location.Start)
            test_message_queue.put(FakeMessage(senv_test_message))

        n_chunks += 1

    assert np.array_equal(data_from_stream.attrs[senv_source_name].value.values,
                          np.concatenate((senv_values_1, senv_values_2)))
    senv_expected_timestamps_1 = np.array([
        senv_timestamp_ns_1, senv_timestamp_ns_1 + senv_time_between_samples,
        senv_timestamp_ns_1 + (2 * senv_time_between_samples)
    ],
                                          dtype=np.dtype('datetime64[ns]'))
    senv_expected_timestamps_2 = np.array([
        senv_timestamp_ns_2, senv_timestamp_ns_2 + senv_time_between_samples,
        senv_timestamp_ns_2 + (2 * senv_time_between_samples)
    ],
                                          dtype=np.dtype('datetime64[ns]'))
    assert np.array_equal(
        data_from_stream.attrs[senv_source_name].value.coords['time'].values,
        np.concatenate((senv_expected_timestamps_1, senv_expected_timestamps_2)))


@pytest.mark.asyncio
async def test_data_stream_returns_data_from_multiple_chopper_messages(queues):
    data_queue, worker_instruction_queue, test_message_queue = queues
    run_info_topic = "fake_topic"
    test_instrument_name = "DATA_STREAM_TEST"

    # The Kafka topics to get metadata from are recorded as "stream" objects in
    # the nexus_structure field of the run start message
    tdct_source_name = "tdct_source"
    tdct_log_name = "tdct_log"
    streams = [
        Stream(f"/entry/{tdct_log_name}", "tdct_topic", tdct_source_name, "tdct")
    ]

    test_stream_args = TEST_STREAM_ARGS.copy()
    test_stream_args["topics"] = None
    n_chunks = 0
    async for data in _data_stream(data_queue,
                                   worker_instruction_queue,
                                   run_info_topic=run_info_topic,
                                   query_consumer=FakeQueryConsumer(
                                       test_instrument_name, streams=streams),
                                   halt_after_n_data_chunks=2,
                                   **test_stream_args,
                                   test_message_queue=test_message_queue):
        data_from_stream = data

        if n_chunks == 0:
            # Fake receiving a Kafka message for each metadata schema
            # Do this after the run start message has been parsed, so that
            # a metadata buffer will have been created for each data source
            # described in the start message.
            tdct_timestamps_1 = np.array([1234, 2345, 3456])  # ns
            tdct_test_message = serialise_tdct(tdct_source_name, tdct_timestamps_1)
            test_message_queue.put(FakeMessage(tdct_test_message))
            tdct_timestamps_2 = np.array([4567, 5678, 6789])  # ns
            tdct_test_message = serialise_tdct(tdct_source_name, tdct_timestamps_2)
            test_message_queue.put(FakeMessage(tdct_test_message))

        n_chunks += 1

    assert np.array_equal(data_from_stream.attrs[tdct_source_name].value.values,
                          np.concatenate((tdct_timestamps_1, tdct_timestamps_2)))


@pytest.mark.asyncio
async def test_data_stream_warns_if_fast_metadata_message_exceeds_buffer(queues):
    data_queue, worker_instruction_queue, test_message_queue = queues
    buffer_size = 2
    run_info_topic = "fake_topic"
    test_instrument_name = "DATA_STREAM_TEST"

    # The Kafka topics to get metadata from are recorded as "stream" objects in
    # the nexus_structure field of the run start message
    senv_source_name = "senv_source"
    senv_log_name = "senv_log"
    streams = [
        Stream(f"/entry/{senv_log_name}", "senv_topic", senv_source_name, "senv",
               "double", "m"),
    ]

    test_stream_args = TEST_STREAM_ARGS.copy()
    test_stream_args["topics"] = None
    test_stream_args["fast_metadata_buffer_size"] = buffer_size
    with pytest.warns(BufferSizeWarning):
        async for _ in _data_stream(data_queue,
                                    worker_instruction_queue,
                                    run_info_topic=run_info_topic,
                                    query_consumer=FakeQueryConsumer(
                                        test_instrument_name, streams=streams),
                                    halt_after_n_warnings=1,
                                    **test_stream_args,
                                    test_message_queue=test_message_queue):
            # Fake receiving a Kafka message for each metadata schema
            # Do this after the run start message has been parsed, so that
            # a metadata buffer will have been created for each data source
            # described in the start message.

            # 3 values but buffer size is only 2!
            senv_values = np.array([26, 127, 52])
            senv_timestamp_ns = 123000  # ns after epoch
            senv_timestamp = datetime.datetime.fromtimestamp(senv_timestamp_ns * 1e-9,
                                                             datetime.timezone.utc)
            senv_time_between_samples = 100  # ns
            senv_test_message = serialise_senv(senv_source_name, -1, senv_timestamp,
                                               senv_time_between_samples, 0,
                                               senv_values, Location.Start)

            test_message_queue.put(FakeMessage(senv_test_message))


@pytest.mark.asyncio
async def test_data_stream_warns_if_single_chopper_message_exceeds_buffer(queues):
    data_queue, worker_instruction_queue, test_message_queue = queues
    buffer_size = 2
    run_info_topic = "fake_topic"
    test_instrument_name = "DATA_STREAM_TEST"

    # The Kafka topics to get metadata from are recorded as "stream" objects in
    # the nexus_structure field of the run start message
    tdct_source_name = "tdct_source"
    tdct_log_name = "tdct_log"
    streams = [
        Stream(f"/entry/{tdct_log_name}", "tdct_topic", tdct_source_name, "tdct")
    ]

    test_stream_args = TEST_STREAM_ARGS.copy()
    test_stream_args["topics"] = None
    test_stream_args["chopper_buffer_size"] = buffer_size
    with pytest.warns(BufferSizeWarning):
        async for _ in _data_stream(data_queue,
                                    worker_instruction_queue,
                                    run_info_topic=run_info_topic,
                                    query_consumer=FakeQueryConsumer(
                                        test_instrument_name, streams=streams),
                                    halt_after_n_warnings=1,
                                    **test_stream_args,
                                    test_message_queue=test_message_queue):
            # Fake receiving a Kafka message for each metadata schema
            # Do this after the run start message has been parsed, so that
            # a metadata buffer will have been created for each data source
            # described in the start message.

            # 3 values but buffer size is only 2!
            tdct_timestamps = np.array([1234, 2345, 3456])  # ns
            tdct_test_message = serialise_tdct(tdct_source_name, tdct_timestamps)

            test_message_queue.put(FakeMessage(tdct_test_message))


@pytest.mark.asyncio
async def test_data_returned_if_multiple_slow_metadata_msgs_exceed_buffer(queues):
    data_queue, worker_instruction_queue, test_message_queue = queues
    run_info_topic = "fake_topic"
    test_instrument_name = "DATA_STREAM_TEST"

    # The Kafka topics to get metadata from are recorded as "stream" objects in
    # the nexus_structure field of the run start message
    f142_source_name = "f142_source"
    f142_log_name = "f142_log"
    streams = [
        Stream(f"/entry/{f142_log_name}", "f142_topic", f142_source_name, "f142",
               "double", "m"),
    ]

    first_f142_value = 26.1236
    f142_timestamp = 123456  # ns after epoch
    first_message = serialise_f142(first_f142_value, f142_source_name, f142_timestamp)
    second_f142_value = 62.721
    second_message = serialise_f142(second_f142_value, f142_source_name, f142_timestamp)

    test_stream_args = TEST_STREAM_ARGS.copy()
    test_stream_args["slow_metadata_buffer_size"] = 1
    test_stream_args["topics"] = None
    n_chunks = 0
    reached_asserts = False
    async for data in _data_stream(data_queue,
                                   worker_instruction_queue,
                                   run_info_topic=run_info_topic,
                                   query_consumer=FakeQueryConsumer(
                                       test_instrument_name, streams=streams),
                                   halt_after_n_data_chunks=3,
                                   **test_stream_args,
                                   test_message_queue=test_message_queue):
        # n_chunks == 0 zeroth chunk contains data
        # from run start message
        if n_chunks == 0:
            test_message_queue.put(FakeMessage(first_message))
            test_message_queue.put(FakeMessage(second_message))
        elif n_chunks == 1:
            # Contains data from first message
            assert isclose(data.attrs[f142_source_name].value.values[0],
                           first_f142_value)
        elif n_chunks == 2:
            # Contains data from second message
            assert isclose(data.attrs[f142_source_name].value.values[0],
                           second_f142_value)
            reached_asserts = True
        n_chunks += 1

    assert reached_asserts


@pytest.mark.asyncio
async def test_data_returned_if_multiple_fast_metadata_msgs_exceed_buffer(queues):
    data_queue, worker_instruction_queue, test_message_queue = queues
    buffer_size = 4
    run_info_topic = "fake_topic"
    test_instrument_name = "DATA_STREAM_TEST"

    # The Kafka topics to get metadata from are recorded as "stream" objects in
    # the nexus_structure field of the run start message
    senv_source_name = "senv_source"
    senv_log_name = "senv_log"
    streams = [
        Stream(f"/entry/{senv_log_name}", "senv_topic", senv_source_name, "senv",
               "double", "m"),
    ]

    first_senv_values = np.array([26, 127, 52])
    second_senv_values = np.array([72, 94, 1])
    senv_timestamp_ns = 123000  # ns after epoch
    senv_timestamp = datetime.datetime.fromtimestamp(senv_timestamp_ns * 1e-9,
                                                     datetime.timezone.utc)
    senv_time_between_samples = 100  # ns
    first_message = serialise_senv(senv_source_name, -1, senv_timestamp,
                                   senv_time_between_samples, 0, first_senv_values,
                                   Location.Start)
    second_message = serialise_senv(senv_source_name, -1, senv_timestamp,
                                    senv_time_between_samples, 0, second_senv_values,
                                    Location.Start)

    test_stream_args = TEST_STREAM_ARGS.copy()
    test_stream_args["topics"] = None
    test_stream_args["fast_metadata_buffer_size"] = buffer_size
    n_chunks = 0
    reached_asserts = False
    async for data in _data_stream(data_queue,
                                   worker_instruction_queue,
                                   run_info_topic=run_info_topic,
                                   query_consumer=FakeQueryConsumer(
                                       test_instrument_name, streams=streams),
                                   halt_after_n_data_chunks=3,
                                   **test_stream_args,
                                   test_message_queue=test_message_queue):
        # n_chunks == 0 zeroth chunk contains data
        # from run start message
        if n_chunks == 0:
            test_message_queue.put(FakeMessage(first_message))
            test_message_queue.put(FakeMessage(second_message))
        elif n_chunks == 1:
            # Contains data from first message
            assert np.array_equal(data.attrs[senv_source_name].value.values,
                                  first_senv_values)
        elif n_chunks == 2:
            # Contains data from second message
            assert np.array_equal(data.attrs[senv_source_name].value.values,
                                  second_senv_values)
            reached_asserts = True
        n_chunks += 1
    assert reached_asserts


@pytest.mark.asyncio
async def test_data_returned_if_multiple_chopper_msgs_exceed_buffer(queues):
    data_queue, worker_instruction_queue, test_message_queue = queues
    buffer_size = 4
    run_info_topic = "fake_topic"
    test_instrument_name = "DATA_STREAM_TEST"

    # The Kafka topics to get metadata from are recorded as "stream" objects in
    # the nexus_structure field of the run start message
    tdct_source_name = "tdct_source"
    tdct_log_name = "tdct_log"
    streams = [
        Stream(f"/entry/{tdct_log_name}", "tdct_topic", tdct_source_name, "tdct")
    ]

    tdct_timestamps_1 = np.array([1234, 2345, 3456])  # ns
    first_tdct_message = serialise_tdct(tdct_source_name, tdct_timestamps_1)
    tdct_timestamps_2 = np.array([4567, 5678, 6789])  # ns
    second_tdct_message = serialise_tdct(tdct_source_name, tdct_timestamps_2)

    test_stream_args = TEST_STREAM_ARGS.copy()
    test_stream_args["topics"] = None
    test_stream_args["chopper_buffer_size"] = buffer_size

    n_chunks = 0
    reached_asserts = False
    async for data in _data_stream(data_queue,
                                   worker_instruction_queue,
                                   run_info_topic=run_info_topic,
                                   query_consumer=FakeQueryConsumer(
                                       test_instrument_name, streams=streams),
                                   **test_stream_args,
                                   halt_after_n_data_chunks=3,
                                   test_message_queue=test_message_queue):
        # n_chunks == 0 zeroth chunk contains data
        # from run start message
        if n_chunks == 0:
            test_message_queue.put(FakeMessage(first_tdct_message))
            test_message_queue.put(FakeMessage(second_tdct_message))
        elif n_chunks == 1:
            # Contains data from first message
            assert np.array_equal(data.attrs[tdct_source_name].value.values,
                                  tdct_timestamps_1)
        elif n_chunks == 2:
            # Contains data from second message
            assert np.array_equal(data.attrs[tdct_source_name].value.values,
                                  tdct_timestamps_2)
            reached_asserts = True
        n_chunks += 1
    assert reached_asserts


@pytest.mark.asyncio
async def test_no_warning_for_missing_datasets_if_group_contains_stream(queues):
    # Create NeXus description for run start message which contains
    # an NXlog which contains no datasets but does have a Stream
    # source for the data
    builder = NexusBuilder()
    test_instrument_name = "DATA_STREAM_TEST"
    builder.add_instrument(test_instrument_name)
    builder.add_log(Log("log", None))
    builder.add_event_data(EventData(None, None, None, None))
    builder.add_stream(Stream("/entry/log"))
    builder.add_stream(Stream("/entry/events_0"))
    nexus_structure = builder.json_string

    data_queue, worker_instruction_queue, test_message_queue = queues
    run_info_topic = "fake_topic"
    reached_assert = False

    with warnings.catch_warnings(record=True) as caught_warnings:
        warnings.simplefilter("always")
        async for _ in _data_stream(data_queue,
                                    worker_instruction_queue,
                                    run_info_topic=run_info_topic,
                                    query_consumer=FakeQueryConsumer(
                                        test_instrument_name,
                                        nexus_structure=nexus_structure),
                                    **TEST_STREAM_ARGS,
                                    halt_after_n_data_chunks=0,
                                    test_message_queue=test_message_queue):
            reached_assert = True
            break
        assert reached_assert
        assert len(
            caught_warnings
        ) == 0, "Expect no 'missing datasets' warning from the NXlog or " \
                "NXevent_data because they each contain a stream which " \
                "will provide the missing data"


@pytest.mark.asyncio
async def test_data_stream_times_out(queues):
    data_queue, worker_instruction_queue, test_message_queue = queues
    run_info_topic = "fake_topic"
    test_instrument_name = "DATA_STREAM_TEST"

    test_stream_args = TEST_STREAM_ARGS.copy()
    test_stream_args["topics"] = None
    # "timeout" is only for use in tests, ensures that if something
    # gets broken then tests will not get stuck in the _data_stream
    # loop indefinitely.
    test_stream_args["timeout"] = 2. * sc.units.s
    timed_out = False
    try:
        async for _ in _data_stream(
                data_queue,
                worker_instruction_queue,
                run_info_topic=run_info_topic,
                query_consumer=FakeQueryConsumer(test_instrument_name),
                **test_stream_args,
                test_message_queue=test_message_queue,
                end_at=StopTime.END_OF_RUN):
            # Do nothing until it times out
            pass
    except TimeoutError:
        timed_out = True
    assert timed_out


@pytest.mark.asyncio
async def test_stream_loop_exits_if_stop_time_and_end_of_partition_reached(queues):
    data_queue, worker_instruction_queue, test_message_queue = queues
    run_info_topic = "fake_topic"
    test_instrument_name = "DATA_STREAM_TEST"

    test_stream_args = TEST_STREAM_ARGS.copy()
    test_stream_args["topics"] = None
    # System time is already after this stop time so the stream will stop
    # as soon as it sees the end of partition or a message with a
    # timestamp after the stop time
    stop_time_in_past = datetime.datetime(2017, 11, 28, 23, 55, 59, 342380)
    n_chunks = 0
    async for _ in _data_stream(data_queue,
                                worker_instruction_queue,
                                run_info_topic=run_info_topic,
                                query_consumer=FakeQueryConsumer(
                                    test_instrument_name, stop_time=stop_time_in_past),
                                **test_stream_args,
                                test_message_queue=test_message_queue,
                                end_at=StopTime.END_OF_RUN):
        if n_chunks == 0:
            # Tell consumer it has reached the end of the partition
            # Consumer will stop, data_consumption_manager will see the
            # consumer has stopped and stop the buffer and the
            # _data_stream will exit.
            # A TimeoutError would occur if the functionality is broken.
            test_message_queue.put(FakeMessage(b"", KafkaError._PARTITION_EOF))
        n_chunks += 1


@pytest.mark.asyncio
async def test_stream_loop_exits_if_stop_time_reached_and_later_message_seen(queues):
    data_queue, worker_instruction_queue, test_message_queue = queues
    run_info_topic = "fake_topic"
    test_instrument_name = "DATA_STREAM_TEST"

    # The Kafka topics to get metadata from are recorded as "stream" objects in
    # the nexus_structure field of the run start message
    f142_source_name = "f142_source"
    f142_log_name = "f142_log"
    streams = [
        Stream(f"/entry/{f142_log_name}", "f142_topic", f142_source_name, "f142",
               "double", "m"),
    ]

    test_stream_args = TEST_STREAM_ARGS.copy()
    test_stream_args["topics"] = None
    # System time is already after this stop time so the stream will stop
    # as soon as it sees the end of partition or a message with a
    # timestamp after the stop time
    stop_time_in_past = datetime.datetime(2017, 11, 28, 23, 55, 59, 342380)
    n_chunks = 0
    async for data in _data_stream(data_queue,
                                   worker_instruction_queue,
                                   run_info_topic=run_info_topic,
                                   query_consumer=FakeQueryConsumer(
                                       test_instrument_name,
                                       stop_time=stop_time_in_past,
                                       streams=streams),
                                   **test_stream_args,
                                   test_message_queue=test_message_queue,
                                   end_at=StopTime.END_OF_RUN):
        if n_chunks == 0:
            # Publish a message with a timestamp before the stop time
            f142_value_1 = 26.1236
            timestamp_before_stop_dt = datetime.datetime(2017, 11, 28, 23, 55, 50, 0)
            # Convert to integer nanoseconds
            # (for timestamp in message payload)
            timestamp_before_stop_ns = int(timestamp_before_stop_dt.timestamp() *
                                           1_000_000_000)
            # Convert to integer milliseconds
            # (for Kafka message header)
            timestamp_before_stop_ms = int(timestamp_before_stop_dt.timestamp() * 1_000)
            f142_test_message = serialise_f142(f142_value_1, f142_source_name,
                                               timestamp_before_stop_ns)
            test_message_queue.put(
                FakeMessage(f142_test_message, timestamp=timestamp_before_stop_ms))
        elif n_chunks == 1:
            # The data from the first message will be returned
            assert np.allclose(data.attrs[f142_source_name].value.values,
                               np.array([f142_value_1]))
            assert np.array_equal(
                data.attrs[f142_source_name].value.coords['time'].values,
                np.array([timestamp_before_stop_ns], dtype=np.dtype('datetime64[ns]')))

            # Publish message with timestamp after stop time, this will trigger
            # the consumer to stop and data_stream to exit.
            # A TimeoutError would occur if the functionality is broken.
            f142_value_2 = 2.725
            timestamp_after_stop_dt = datetime.datetime(2017, 11, 28, 23, 56, 50, 0)
            timestamp_after_stop_ns = int(timestamp_after_stop_dt.timestamp() *
                                          1_000_000_000)
            timestamp_after_stop_ms = int(timestamp_after_stop_dt.timestamp() * 1_000)
            f142_test_message = serialise_f142(f142_value_2, f142_source_name,
                                               timestamp_after_stop_ns)
            test_message_queue.put(
                FakeMessage(f142_test_message, timestamp=timestamp_after_stop_ms))

        n_chunks += 1



import scipp as sc
import scippnexus as nexus
import scippneutron as scn


def test_can_load_nxdetector_from_bigfake():
    with nexus.File(scn.data.bigfake()) as f:
        da = f['entry/instrument/detector_1'][...]
        assert da.sizes == {'dim_0': 300, 'dim_1': 300}


def test_can_load_nxdetector_from_PG3():
    import scippneutron as scn
    with nexus.File(scn.data.get_path('PG3_4844_event.nxs')) as f:
        det = f['entry/instrument/bank24']
        da = det[...]
        assert da.sizes == {'x_pixel_offset': 154, 'y_pixel_offset': 7}
        assert 'detector_number' not in da.coords
        assert da.coords['pixel_id'].sizes == da.sizes
        assert da.coords['distance'].sizes == da.sizes
        assert da.coords['polar_angle'].sizes == da.sizes
        assert da.coords['azimuthal_angle'].sizes == da.sizes
        assert da.coords['x_pixel_offset'].sizes == {'x_pixel_offset': 154}
        assert da.coords['y_pixel_offset'].sizes == {'y_pixel_offset': 7}
        # local_name is an example of a dataset with shape=[1] that is treated as scalar
        assert da.coords['local_name'].sizes == {}
        # Extra scalar fields not in underlying NXevent_data
        del da.coords['local_name']
        del da.coords['total_counts']
        assert sc.identical(da.sum(), det.events[()].sum())  # no event lost in binning



import sys
from typing import List, Union
from packaging.version import parse, Version, InvalidVersion
import requests
import argparse


def _get_releases(repo: str, organization: str = 'scipp') -> List[Version]:
    """Return reversed sorted list of release tag names."""
    r = requests.get(f'https://api.github.com/repos/{organization}/{repo}/releases')
    if r.status_code != 200:
        return []
    data = r.json()
    return sorted([parse(e['tag_name']) for e in data if not e['draft']], reverse=True)


class VersionInfo:

    def __init__(self, repo: str, organization: str = 'scipp'):
        self._releases = _get_releases(repo=repo, organization=organization)

    def _to_version(self, version) -> Version:
        if isinstance(version, str):
            try:
                return parse(version)
            except InvalidVersion:
                # When not building for a tagged release we may get, e.g., 'main'.
                # Pretend this means the current latest release.
                return self._releases[0]
        return version

    def minor_releases(self, first: str = '0.1') -> List[str]:
        """Return set of minor releases in the form '1.2'.

        `first` gives the first release to be included. By default, '0.0' releases are
        not included.
        """
        first = parse(first)
        releases = [r for r in self._releases if r >= first]
        releases = sorted(set((r.major, r.minor) for r in releases), reverse=True)
        return [f'{major}.{minor}' for major, minor in releases]

    def is_latest(self, version: Union[str, Version]) -> bool:
        """Return True if `version` has the same or larger major and minor as the
        latest release.
        """
        version = self._to_version(version)
        latest = self._releases[0]
        return (latest.major, latest.minor) <= (version.major, version.minor)

    def is_new(self, version: str) -> bool:
        """Return True if `version` is a new major or minor release."""
        version = self._to_version(version)
        releases = [r for r in self._releases if r != version]
        latest = releases[0]
        return (latest.major, latest.minor) < (version.major, version.minor)

    def target(self, version: Union[str, Version]) -> str:
        version = self._to_version(version)
        if self.is_latest(version):
            return ''
        else:
            return f'release/{version.major}.{version.minor}'

    def replaced(self, version: str) -> Version:
        version = self._to_version(version)
        # If we release 1.2 we want to find 1.1
        for release in self._releases:
            if (release.major, release.minor) < (version.major, version.minor):
                return release


def main(repo: str, action: str, version: str) -> int:
    info = VersionInfo(repo=repo)
    if action == 'is-latest':
        print(info.is_latest(version))
    elif action == 'is-new':
        print(info.is_new(version))
    elif action == 'get-replaced':
        print(info.replaced(version))
    elif action == 'get-target':
        print(info.target(version))
    return 0


if __name__ == '__main__':
    parser = argparse.ArgumentParser(description='Process some integers.')
    parser.add_argument('--repo', dest='repo', required=True, help='Repository name')
    parser.add_argument('--action',
                        choices=['is-latest', 'is-new', 'get-replaced', 'get-target'],
                        required=True,
                        help='Action to perform: Check whether this major or minor '
                        'release exists or is new (is-latest), check whether this is a '
                        'new major or minor release (is-new), get the version this is '
                        'replacing (get-replaced), get the target folder for '
                        'publishing the docs (get-target). In all cases the '
                        'patch/micro version is ignored.')
    parser.add_argument('--version',
                        dest='version',
                        required=True,
                        help='Version the action refers to')

    args = parser.parse_args()
    sys.exit(main(**vars(args)))



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2022 Scipp contributors (https://github.com/scipp)
# @author Matthew Jones
from __future__ import annotations
from typing import Tuple, Dict, List, Any, Union
from dataclasses import dataclass
import numpy as np
from scippnexus.typing import H5Group

_nexus_class = "NX_class"
_nexus_units = "units"
_nexus_name = "name"
_nexus_path = "path"
_nexus_values = "values"
_nexus_dataset = "dataset"
_nexus_group = "group"
_nexus_children = "children"
_nexus_link = "link"
_nexus_stream = "stream"

_filewriter_to_supported_numpy_dtype = {
    "float32": np.float32,
    "float64": np.float64,
    "float": np.float32,
    "double": np.float64,
    "int8": np.int32,
    "int16": np.int32,
    "int32": np.int32,
    "int64": np.int64,
    "uint8": np.int32,
    "uint16": np.int32,
    "uint32": np.int32,
    "uint64": np.int64,
    "string": np.str_
}

numpy_to_filewriter_type = {
    np.float32: "float32",
    np.float64: "float64",
    np.int8: "int8",
    np.int16: "int16",
    np.int32: "int32",
    np.int64: "int64",
    np.uint8: "uint8",
    np.uint16: "uint16",
    np.uint32: "uint32",
    np.uint64: "uint64",
    np.str_: "string",
    np.object_: "string"
}


class MissingAttribute(Exception):
    pass


def make_json_attr(name: str, value) -> dict:
    if isinstance(value, (str, bytes)):
        attr_info = {"string_size": len(value), "type": "string"}
    elif isinstance(value, float):
        attr_info = {"size": 1, "type": "float64"}
    elif isinstance(value, int):
        attr_info = {"size": 1, "type": "int64"}
    elif isinstance(value, list):
        attr_info = {"size": len(value), "type": "string"}
    else:
        attr_info = {
            "size": value.shape,
            "type": numpy_to_filewriter_type[value.dtype.type]
        }
    name_and_value = {"name": name, "values": value}
    return {**attr_info, **name_and_value}


def make_json_dataset(name: str, data) -> dict:
    if isinstance(data, (str, bytes)):
        dataset_info = {"string_size": len(data), "type": "string"}
    elif isinstance(data, float):
        dataset_info = {"size": 1, "type": "float64"}
    elif isinstance(data, int):
        dataset_info = {"size": 1, "type": "int32"}
    else:
        dataset_info = {
            "size": data.shape,
            "type": numpy_to_filewriter_type[data.dtype.type]
        }

    return {
        "type": "dataset",
        "name": name,
        "values": data,
        "dataset": dataset_info,
        "attributes": []
    }


def _get_attribute_value(element: Dict,
                         attribute_name: str) -> Union[str, float, int, List]:
    """
    attributes can be a dictionary of key-value pairs, or an array
    of dictionaries with key, value, type, etc
    """
    try:
        attributes = element["attributes"]
        try:
            return attributes[attribute_name]
        except TypeError:
            for attribute in attributes:
                if attribute[_nexus_name] == attribute_name:
                    return attribute[_nexus_values]
    except KeyError:
        pass
    raise MissingAttribute


class _Node(dict):

    def __init__(self, parent: dict, name: str, file: dict, group: dict):
        super().__init__(**group)
        self.parent = parent
        self.name = name
        self.file = file


def _visit_nodes(root: Dict, group: Dict, nx_class_names: Tuple[str, ...],
                 groups_with_requested_nx_class: Dict[str,
                                                      List[H5Group]], path: List[str]):
    try:
        for child in group[_nexus_children]:
            try:
                path.append(child[_nexus_name])
            except KeyError:
                # If the object doesn't have a name it can't be a NeXus
                # class we are looking for, nor can it be a group
                # containing a NeXus class we are looking for, so skip to
                # next object
                continue
            try:
                nx_class = _get_attribute_value(child, _nexus_class)
                if nx_class in nx_class_names:
                    groups_with_requested_nx_class[nx_class].append(
                        _Node(group=child, parent=group, name="/".join(path),
                              file=root))
            except MissingAttribute:
                # It may be a group but not an NX_class,
                # that's fine, continue to its children
                pass
            _visit_nodes(root, child, nx_class_names, groups_with_requested_nx_class,
                         path)
            path.pop(-1)
    except KeyError:
        pass


def contains_stream(group: Dict) -> bool:
    """
    Return True if the group contains a stream object
    """
    if not isinstance(group, JSONGroup):
        return False
    try:
        for child in group._node[_nexus_children]:
            try:
                if child["type"] == _nexus_stream:
                    return True
            except KeyError:
                # "type" field ought to exist, but if it does
                # not then assume it is not a stream
                pass
    except KeyError:
        # "children" field may be missing, that is okay
        # but means this this group cannot contain a stream
        pass
    return False


def _find_by_type(type_name: str, root: Dict) -> List[H5Group]:
    """
    Finds objects with the requested "type" value
    Returns a list of objects with requested type
    """

    def _visit_nodes_for_type(obj: Dict, requested_type: str,
                              objects_found: List[H5Group]):
        try:
            for child in obj[_nexus_children]:
                if child["type"] == requested_type:
                    objects_found.append(
                        _Node(group=child,
                              parent=obj,
                              name="",
                              file={_nexus_children: [obj]}))
                _visit_nodes_for_type(child, requested_type, objects_found)
        except KeyError:
            # If this object does not have "children" array then go to next
            pass

    objects_with_requested_type: List[H5Group] = []
    _visit_nodes_for_type(root, type_name, objects_with_requested_type)

    return objects_with_requested_type


class JSONTypeStringID():

    def get_cset(self):
        import h5py
        return h5py.h5t.CSET_UTF8


class JSONAttrID:

    def __init__(self):
        pass

    def get_type(self):
        return JSONTypeStringID()


class JSONAttributeManager:

    def __init__(self, node: dict):
        self._node = node

    def __contains__(self, name):
        try:
            self[name]
        except MissingAttribute:
            return False
        return True

    def __getitem__(self, name):
        return _get_attribute_value(self._node, name)

    def __setitem__(self, name, value):
        if name in self:
            raise NotImplementedError("Replacing existing item not implemented yet.")
        attr = make_json_attr(name, value)
        self._node['attributes'].append(attr)
        return self[name]

    def __iter__(self):
        if (attrs := self._node.get('attributes')) is not None:
            if isinstance(attrs, dict):
                yield from attrs
            else:
                for item in attrs:
                    yield item[_nexus_name]

    def get(self, name: str, default=None):
        return self[name] if name in self else default

    def get_id(self, name) -> JSONAttrID:
        # TODO This is a hack that only works since this is used only for a single
        # purpose by scippnexus.NXobject
        return JSONAttrID()


class JSONNode:

    def __init__(self, node: dict, *, parent=None):
        self._file = parent.file if parent is not None else self
        self._parent = self if parent is None else parent
        self._node = node
        if parent is None or parent.name == '/':
            self._name = f'/{self._node.get(_nexus_name, "")}'
        else:
            self._name = f'{parent.name}/{self._node[_nexus_name]}'

    @property
    def attrs(self) -> JSONAttributeManager:
        return JSONAttributeManager(self._node)

    @property
    def name(self) -> str:
        return self._name

    @property
    def file(self):
        return self._file

    @property
    def parent(self):
        return self._parent


class JSONDataset(JSONNode):

    @property
    def dtype(self) -> str:
        try:
            dtype = self._node[_nexus_dataset]["type"]
        except KeyError:
            dtype = self._node[_nexus_dataset]["dtype"]
        if dtype == 'string':
            return np.dtype(str)
        return np.dtype(dtype)

    @property
    def ndim(self) -> int:
        return len(self.shape)

    @property
    def shape(self):
        return np.asarray(self._node[_nexus_values]).shape

    def __getitem__(self, index):
        return np.asarray(self._node[_nexus_values])[index]

    def read_direct(self, buf, source_sel):
        buf[...] = self[source_sel]

    def asstr(self, **ignored):
        return self


class JSONGroup(JSONNode):

    def __contains__(self, name: str) -> bool:
        try:
            self[name]
            return True
        except KeyError:
            return False

    def keys(self) -> List[str]:
        children = self._node[_nexus_children]
        return [child[_nexus_name] for child in children if not contains_stream(child)]

    def _as_group_or_dataset(self, item, parent):
        if item['type'] == _nexus_group:
            return JSONGroup(item, parent=parent)
        else:
            return JSONDataset(item, parent=parent)

    def __getitem__(self, name: str) -> Union[JSONDataset, JSONGroup]:
        if name.startswith('/') and name.count('/') == 1:
            parent = self.file
        elif '/' in name:
            parent = self['/'.join(name.split('/')[:-1])]
        else:
            parent = self

        for child in parent._node[_nexus_children]:
            if child.get(_nexus_name) != name.split('/')[-1]:
                continue
            if child.get('type') == _nexus_link:
                return self[child["target"]]
            if child.get('type') in (_nexus_dataset, _nexus_group):
                return self._as_group_or_dataset(child, parent)

        raise KeyError(f"Unable to open object (object '{name}' doesn't exist)")

    def __iter__(self):
        yield from self.keys()

    def visititems(self, callable):

        def skip(node):
            return node['type'] == _nexus_link or contains_stream(self)

        children = [
            child[_nexus_name] for child in self._node[_nexus_children]
            if not skip(child)
        ]
        for key in children:
            item = self[key]
            callable(key, item)
            if isinstance(item, JSONGroup):
                item.visititems(callable)

    def create_dataset(self, name: str, data) -> JSONDataset:
        if name in self:
            raise NotImplementedError("Replacing existing item not implemented yet.")
        dataset = make_json_dataset(name, data)
        self._node[_nexus_children].append(dataset)
        return self[name]

    def create_group(self, name: str) -> JSONGroup:
        if name in self:
            raise NotImplementedError("Replacing existing item not implemented yet.")
        group = {"type": "group", "name": name, "children": [], "attributes": []}
        self._node[_nexus_children].append(group)
        return self[name]


@dataclass
class StreamInfo:
    topic: str
    flatbuffer_id: str
    source_name: str
    dtype: Any
    unit: str


def get_streams_info(root: Dict) -> List[StreamInfo]:
    found_streams = _find_by_type(_nexus_stream, root)
    streams = []
    for stream in found_streams:
        try:
            dtype = _filewriter_to_supported_numpy_dtype[stream["stream"]["dtype"]]
        except KeyError:
            try:
                dtype = _filewriter_to_supported_numpy_dtype[stream["stream"]["type"]]
            except KeyError:
                dtype = None

        units = "dimensionless"
        try:
            units = _get_attribute_value(stream.parent, _nexus_units)
        except MissingAttribute:
            pass

        streams.append(
            StreamInfo(stream["stream"]["topic"], stream["stream"]["writer_module"],
                       stream["stream"]["source"], dtype, units))
    return streams



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2022 Scipp contributors (https://github.com/scipp)
from dataclasses import dataclass
from ..file_loading.load_nexus import StreamInfo
from typing import Optional, List
from enum import Enum
import multiprocessing as mp
from ._consumer_type import ConsumerType
from ._consumer import (start_consumers, stop_consumers, create_consumers,
                        all_consumers_stopped)
from ._data_buffer import StreamedDataBuffer
from queue import Empty as QueueEmpty


class InstructionType(Enum):
    STOP_NOW = 1
    UPDATE_STOP_TIME = 2


@dataclass(frozen=True)
class ManagerInstruction:
    type: InstructionType
    stop_time_ms: Optional[int] = None  # milliseconds from unix epoch


def data_consumption_manager(start_time_ms: int, stop_time_ms: Optional[int],
                             run_id: str, topics: List[str], kafka_broker: str,
                             consumer_type: ConsumerType,
                             stream_info: Optional[List[StreamInfo]], interval_s: float,
                             event_buffer_size: int, slow_metadata_buffer_size: int,
                             fast_metadata_buffer_size: int, chopper_buffer_size: int,
                             worker_instruction_queue: mp.Queue, data_queue: mp.Queue,
                             test_message_queue: Optional[mp.Queue]):
    """
    Starts and stops buffers and data consumers which collect data and
    send them back to the main process via a queue.

    All input args must be mp.Queue or pickleable as this function is launched
    as a multiprocessing.Process.
    """
    buffer = StreamedDataBuffer(data_queue, event_buffer_size,
                                slow_metadata_buffer_size, fast_metadata_buffer_size,
                                chopper_buffer_size, interval_s, run_id)

    if stream_info is not None:
        buffer.init_metadata_buffers(stream_info)

    consumers = create_consumers(start_time_ms, stop_time_ms, set(topics), kafka_broker,
                                 consumer_type, buffer.new_data, test_message_queue)

    start_consumers(consumers)
    buffer.start()

    while not all_consumers_stopped(consumers):
        try:
            instruction = worker_instruction_queue.get(timeout=.5)
            if instruction.type == InstructionType.STOP_NOW:
                stop_consumers(consumers)
            elif instruction.type == InstructionType.UPDATE_STOP_TIME:
                for consumer in consumers:
                    consumer.update_stop_time(instruction.stop_time_ms)
        except QueueEmpty:
            pass
        except (ValueError, OSError):
            # Queue has been closed, stop worker
            stop_consumers(consumers)

    buffer.stop()



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2022 Scipp contributors (https://github.com/scipp)
import pathlib
from typing import List, Type, Union, Callable
import warnings

from dateutil.parser import parse as parse_date
import numpy as np
import pytest
import scipp as sc

import scippneutron
from scippneutron.file_loading.load_nexus import _load_nexus_json
from .nexus_helpers import (
    NexusBuilder,
    EventData,
    Detector,
    Log,
    Sample,
    Source,
    Transformation,
    TransformationType,
    Link,
    Monitor,
    Chopper,
    in_memory_hdf5_file_with_two_nxentry,
)

# representative sample of UTF-8 test strings from
# https://www.w3.org/2001/06/utf-8-test/UTF-8-demo.html
UTF8_TEST_STRINGS = (
    "∮ E⋅da = Q,  n → ∞, ∑ f(i) = ∏ g(i), ∀x∈ℝ: ⌈x⌉ = −⌊−x⌋, α ∧ ¬β = ¬(¬α ∨ β)",
    "2H₂ + O₂ ⇌ 2H₂O, R = 4.7 kΩ, ⌀ 200 mm",
    "Σὲ γνωρίζω ἀπὸ τὴν κόψη",
)


def _timestamp(date: str):
    return parse_date(date).timestamp()


def test_raises_exception_if_multiple_nxentry_in_file():
    with in_memory_hdf5_file_with_two_nxentry() as nexus_file:
        with pytest.raises(RuntimeError):
            scippneutron.load_nexus(nexus_file)


def test_no_exception_if_single_nxentry_found_below_root():
    with in_memory_hdf5_file_with_two_nxentry() as nexus_file:
        # There are 2 NXentry in the file, but root is used
        # to specify which to load data from
        assert scippneutron.load_nexus(nexus_file, root='/entry_1') is None


def load_from_nexus(builder: NexusBuilder, *args, **kwargs)\
        -> Union[sc.Dataset, sc.DataArray, None]:
    with builder.file() as nexus_file:
        return scippneutron.load_nexus(nexus_file, *args, **kwargs)


def load_from_json(builder: NexusBuilder, *args, **kwargs)\
        -> Union[sc.Dataset, sc.DataArray, None]:
    loaded_data, _ = _load_nexus_json(builder.json_string, *args, **kwargs)
    return loaded_data


@pytest.fixture(params=[load_from_nexus, load_from_json])
def load_function(request) -> Callable:
    """
    Each test with this fixture is executed with load_nexus_json
    loading JSON output from the NexusBuilder, and with load_nexus
    loading in-memory NeXus output from the NexusBuilder
    """
    return request.param


@pytest.mark.parametrize('path_type', (str, pathlib.Path))
def test_loads_from_file(path_type):
    with warnings.catch_warnings():
        warnings.filterwarnings("ignore", message='Skipped load', category=UserWarning)
        da = scippneutron.load_nexus(path_type(scippneutron.data.bigfake()))
    assert da.sizes == {'detector_id': 90000, 'tof': 1}
    assert 'tof' in da.coords
    assert 'position' in da.coords


def test_no_exception_if_single_nxentry_in_file(load_function: Callable):
    builder = NexusBuilder()
    assert load_function(builder) is None


def test_loads_data_from_single_event_data_group(load_function: Callable):
    event_time_offsets = np.array([456, 743, 347, 345, 632])
    event_data = EventData(
        event_id=np.array([1, 2, 3, 1, 3]),
        event_time_offset=event_time_offsets,
        event_time_zero=np.array([
            1600766730000000000, 1600766731000000000, 1600766732000000000,
            1600766733000000000
        ]),
        event_index=np.array([0, 3, 3, 5]),
    )

    builder = NexusBuilder()
    builder.add_event_data(event_data)

    loaded_data = load_function(builder)

    assert loaded_data.bins.constituents['data'].unit == 'counts'

    # Expect time of flight to match the values in the
    # event_time_offset dataset
    # May be reordered due to binning (hence np.sort)
    assert sc.identical(
        sc.sort(loaded_data.bins.concat('detector_id').values[0].coords['tof'],
                key="event"),
        sc.sort(sc.array(dims=["event"], values=event_time_offsets, unit=sc.units.ns),
                key="event"))

    counts_on_detectors = loaded_data.bins.sum()
    # No detector_number dataset in file so expect detector_id to be
    # binned according to whatever detector_ids are present in event_id
    # dataset: 2 on det 1, 1 on det 2, 2 on det 3
    expected_counts = np.array([[2], [1], [2]])
    assert sc.identical(
        counts_on_detectors.data,
        sc.array(dims=['detector_id', 'tof'],
                 unit='counts',
                 dtype='float32',
                 values=expected_counts,
                 variances=expected_counts))
    expected_detector_ids = np.array([1, 2, 3])
    assert np.array_equal(loaded_data.coords['detector_id'].values,
                          expected_detector_ids)


@pytest.mark.parametrize("unit,multiplier",
                         (("ns", 1), ("us", 10**3), ("ms", 10**6), ("s", 10**9)))
def test_loads_pulse_times_from_single_event_with_different_units(
        load_function: Callable, unit: str, multiplier: float):

    offsets = np.array([12, 34, 56, 78])
    zeros = np.array([12., 34., 56., 78.], dtype="float64")
    event_data = EventData(
        event_id=np.array([1, 2, 3, 4]),
        event_time_offset=offsets,
        event_time_zero=zeros,
        event_index=np.array([0, 3, 3, 4]),
        event_time_zero_unit=unit,
    )

    builder = NexusBuilder()
    builder.add_detector(
        Detector(detector_numbers=np.array([1, 2, 3, 4]), event_data=event_data))

    loaded_data = load_function(builder)

    for event, pulse_time in enumerate([12, 12, 12, 56]):
        _time = np.array("1970-01-01").astype("datetime64[ns]") \
                + np.array(pulse_time).astype("timedelta64[ns]") * multiplier

        # Allow 1ns difference for rounding errors between different routes
        assert all(
            np.abs(loaded_data.values[event].coords['pulse_time'].values -
                   _time) <= np.array(1).astype("timedelta64[ns]"))


@pytest.mark.parametrize("time_zero_offset,time_zero,time_zero_unit,expected_time", (
    ("1980-01-01T00:00:00.0", 30, "s", "1980-01-01T00:00:30.0"),
    ("1990-01-01T00:00:00.0", 5000, "ms", "1990-01-01T00:00:05.0"),
    ("2000-01-01T00:00:00.0", 3 * 10**6, "us", "2000-01-01T00:00:03.0"),
    ("2010-01-01T00:00:00.0", 12, "hour", "2010-01-01T12:00:00.0"),
))
def test_loads_pulse_times_with_combinations_of_offset_and_units(
        load_function: Callable, time_zero_offset: str, time_zero: float,
        time_zero_unit: str, expected_time: str):

    offsets = np.array([0])
    zeros = np.array([time_zero], dtype="float64")
    event_data = EventData(
        event_id=np.array([0]),
        event_time_offset=offsets,
        event_time_zero_offset=time_zero_offset,
        event_time_zero=zeros,
        event_index=np.array([0]),
        event_time_zero_unit=time_zero_unit,
    )

    builder = NexusBuilder()
    builder.add_detector(Detector(detector_numbers=np.array([0]),
                                  event_data=event_data))

    loaded_data = load_function(builder)

    _time = np.array(expected_time).astype("datetime64[ns]")

    # Allow 1ns difference for rounding errors between different routes
    assert np.abs(loaded_data.values[0].coords['pulse_time'].values[0] -
                  _time) <= np.array(1).astype("timedelta64[ns]")


def test_does_not_load_events_if_index_not_ordered(load_function: Callable):
    event_data_1 = EventData(
        event_id=np.array([0, 1]),
        event_time_offset=np.array([0, 1]),
        event_time_zero=np.array([0, 1]),
        event_index=np.array([2, 0]),
    )

    builder = NexusBuilder()
    builder.add_detector(
        Detector(detector_numbers=np.array([0, 1]), event_data=event_data_1))

    with pytest.warns(UserWarning, match="Invalid index in NXevent_data at "):
        load_function(builder)


def test_loads_pulse_times_from_multiple_event_data_groups(load_function: Callable):
    offsets = np.array([0, 0, 0, 0])

    zeros_1 = np.array([12 * 10**9, 34 * 10**9, 56 * 10**9, 78 * 10**9])
    zeros_2 = np.array([87 * 10**9, 65 * 10**9, 43 * 10**9, 21 * 10**9])

    event_data_1 = EventData(
        event_id=np.array([0, 1, 2, 3]),
        event_time_offset=offsets,
        event_time_zero=zeros_1,
        event_index=np.array([0, 3, 3, 4]),
    )
    event_data_2 = EventData(
        event_id=np.array([4, 5, 6, 7]),
        event_time_offset=offsets,
        event_time_zero=zeros_2,
        event_index=np.array([0, 3, 3, 4]),
    )

    builder = NexusBuilder()
    builder.add_detector(
        Detector(detector_numbers=np.array([0, 1, 2, 3]), event_data=event_data_1))
    builder.add_detector(
        Detector(detector_numbers=np.array([4, 5, 6, 7]), event_data=event_data_2))

    loaded_data = load_function(builder)

    for event, pulse_time in enumerate([12, 12, 12, 56, 87, 87, 87, 43]):
        _time = np.array("1970-01-01").astype("datetime64[ns]") \
                + np.array(pulse_time).astype("timedelta64[s]")

        assert sc.identical(
            loaded_data.values[event].coords['pulse_time'],
            sc.array(dims=["event"],
                     values=[_time],
                     unit=sc.units.ns,
                     dtype=sc.DType.datetime64))


def test_loads_data_from_multiple_event_data_groups(load_function: Callable):
    pulse_times = np.array([
        1600766730000000000, 1600766731000000000, 1600766732000000000,
        1600766733000000000
    ])
    event_time_offsets_1 = np.array([456, 743, 347, 345, 632])
    event_data_1 = EventData(
        event_id=np.array([1, 2, 3, 1, 3]),
        event_time_offset=event_time_offsets_1,
        event_time_zero=pulse_times,
        event_index=np.array([0, 3, 3, 5]),
    )
    detector_1_ids = np.array([0, 1, 2, 3])
    event_time_offsets_2 = np.array([682, 237, 941, 162, 52])
    event_data_2 = EventData(
        event_id=np.array([4, 5, 6, 4, 6]),
        event_time_offset=event_time_offsets_2,
        event_time_zero=pulse_times,
        event_index=np.array([0, 3, 3, 5]),
    )
    detector_2_ids = np.array([4, 5, 6, 7])

    builder = NexusBuilder()
    builder.add_detector(Detector(detector_1_ids, event_data_1))
    builder.add_detector(Detector(detector_2_ids, event_data_2))

    loaded_data = load_function(builder)

    # Expect time of flight to match the values in the
    # event_time_offset datasets
    # May be reordered due to binning (hence np.sort)
    assert np.array_equal(
        np.sort(loaded_data.bins.concat('detector_id').values[0].coords['tof'].values),
        np.sort(np.concatenate((event_time_offsets_1, event_time_offsets_2))))

    counts_on_detectors = loaded_data.bins.sum()
    # There are detector_number datasets in the NXdetector for each
    # NXevent_data, these are used for detector_id binning
    expected_counts = np.array([[0], [2], [1], [2], [2], [1], [2], [0]])
    assert np.array_equal(counts_on_detectors.data.values, expected_counts)
    expected_detector_ids = np.concatenate((detector_1_ids, detector_2_ids))
    assert np.array_equal(loaded_data.coords['detector_id'].values,
                          expected_detector_ids)


def test_skips_event_data_group_with_non_integer_event_ids(load_function: Callable):
    event_time_offsets = np.array([456, 743, 347, 345, 632])
    event_data = EventData(
        event_id=np.array([1.1, 2.2, 3.3, 1.1, 3.1]),
        event_time_offset=event_time_offsets,
        event_time_zero=np.array([
            1600766730000000000, 1600766731000000000, 1600766732000000000,
            1600766733000000000
        ]),
        event_index=np.array([0, 3, 3, 5]),
    )

    builder = NexusBuilder()
    builder.add_event_data(event_data)

    with pytest.warns(UserWarning):
        loaded_data = load_function(builder)

    assert loaded_data is None, "Expected no data to be loaded as " \
                                "event data has non integer event ids"


def test_skips_event_data_group_with_non_integer_detector_numbers(
        load_function: Callable):
    event_time_offsets = np.array([456, 743, 347, 345, 632])
    event_data = EventData(
        event_id=np.array([1, 2, 3, 1, 3]),
        event_time_offset=event_time_offsets,
        event_time_zero=np.array([
            1600766730000000000, 1600766731000000000, 1600766732000000000,
            1600766733000000000
        ]),
        event_index=np.array([0, 3, 3, 5]),
    )
    detector_numbers = np.array([0.1, 1.2, 2.3, 3.4])

    builder = NexusBuilder()
    builder.add_detector(Detector(detector_numbers, event_data))

    with pytest.warns(UserWarning):
        loaded_data = load_function(builder)

    assert loaded_data is None, "Expected no data to be loaded as " \
                                "detector has non integer detector numbers"


def test_skips_data_with_event_id_and_detector_number_type_unequal(
        load_function: Callable):
    event_time_offsets = np.array([456, 743, 347, 345, 632])
    event_data = EventData(
        event_id=np.array([1, 2, 3, 1, 3], dtype=np.int64),
        event_time_offset=event_time_offsets,
        event_time_zero=np.array([
            1600766730000000000, 1600766731000000000, 1600766732000000000,
            1600766733000000000
        ]),
        event_index=np.array([0, 3, 3, 5]),
    )
    detector_numbers = np.array([0, 1, 2, 3], dtype=np.int32)

    builder = NexusBuilder()
    builder.add_detector(Detector(detector_numbers, event_data))

    with pytest.warns(UserWarning):
        loaded_data = load_function(builder)

    assert loaded_data is None, "Expected no data to be loaded as event " \
                                "ids and detector numbers are of " \
                                "different types"


def test_loads_data_from_single_log_with_no_units(load_function: Callable):
    values = np.array([1, 2, 3])
    times = np.array([4, 5, 6])
    name = "test_log"
    builder = NexusBuilder()
    builder.add_log(Log(name, values, times))

    loaded_data = load_function(builder)

    # Expect a sc.Dataset with log names as keys
    assert np.array_equal(loaded_data[name].data.values.values, values)


def test_loads_data_from_single_log_with_units(load_function: Callable):
    values = np.array([1.1, 2.2, 3.3])
    times = np.array([4.4, 5.5, 6.6])
    name = "test_log"
    builder = NexusBuilder()
    builder.add_log(Log(name, values, times, value_units="m", time_units="s"))

    loaded_data = load_function(builder)

    # Expect a sc.Dataset with log names as keys
    assert np.allclose(loaded_data[name].data.values.values, values)
    assert loaded_data[name].data.values.unit == sc.units.m


def test_loads_data_from_multiple_logs(load_function: Callable):
    builder = NexusBuilder()
    log_1 = Log("test_log", np.array([1.1, 2.2, 3.3]), np.array([4.4, 5.5, 6.6]))
    log_2 = Log("test_log_2", np.array([123, 253, 756]), np.array([246, 1235, 2369]))
    builder.add_log(log_1)
    builder.add_log(log_2)

    loaded_data = load_function(builder)

    # Expect a sc.Dataset with log names as keys
    assert np.allclose(loaded_data[log_1.name].data.values.values, log_1.value)
    assert np.array_equal(loaded_data[log_2.name].data.values.values, log_2.value)


def test_loads_logs_with_non_supported_int_types(load_function: Callable):
    builder = NexusBuilder()
    log_int8 = Log("test_log_int8",
                   np.array([1, 2, 3]).astype(np.int8), np.array([4.4, 5.5, 6.6]))
    log_int16 = Log("test_log_int16",
                    np.array([123, 253, 756]).astype(np.int16),
                    np.array([246, 1235, 2369]))
    log_uint8 = Log("test_log_uint8",
                    np.array([1, 2, 3]).astype(np.uint8), np.array([4.4, 5.5, 6.6]))
    log_uint16 = Log("test_log_uint16",
                     np.array([123, 253, 756]).astype(np.uint16),
                     np.array([246, 1235, 2369]))
    logs = (log_int8, log_int16, log_uint8, log_uint16)
    for log in logs:
        builder.add_log(log)

    loaded_data = load_function(builder)

    # Expect a sc.Dataset with log names as keys
    for log in logs:
        assert np.allclose(loaded_data[log.name].data.values.values, log.value)


def test_loads_multidimensional_log(load_function: Callable):
    multidim_values = np.array([[1, 2, 3], [1, 2, 3]])
    times = np.array([4, 5])
    name = "test_log"
    builder = NexusBuilder()
    builder.add_log(Log(name, multidim_values, times, value_units=''))

    loaded_data = load_function(builder)

    log = loaded_data[name].data.value
    expected = sc.DataArray(sc.array(dims=['time', 'dim_1'], values=multidim_values))
    expected.coords['time'] = sc.epoch(unit='s') + sc.array(
        dims=['time'], values=times, unit='s')
    assert sc.identical(log, expected)


def test_skips_log_with_no_value_dataset(load_function: Callable):
    name = "test_log"
    builder = NexusBuilder()
    builder.add_log(Log(name, None, np.array([4, 5, 6])))

    with pytest.warns(UserWarning):
        loaded_data = load_function(builder)

    assert loaded_data is None


def test_loads_log_with_empty_value_and_time_datasets(load_function: Callable):
    empty_values = np.array([]).astype(np.int32)
    empty_times = np.array([]).astype(np.int32)
    name = "test_log"
    builder = NexusBuilder()
    builder.add_log(Log(name, empty_values, empty_times))

    loaded_data = load_function(builder)

    assert loaded_data[name].data.values.sizes == {'time': 0}


def test_warns_given_log_with_mismatched_value_and_time(load_function: Callable):
    values = np.array([1, 2, 3]).astype(np.int32)
    # Note that if times exceeds length by 1 it is loaded as bin edges. It is unclear
    # if this is considered valid Nexus.
    times = np.array([1, 2, 3, 4, 5]).astype(np.int32)
    name = "test_log"
    builder = NexusBuilder()
    builder.add_log(Log(name, values, times))

    with pytest.warns(UserWarning):
        loaded_data = load_function(builder)

    assert 'time' not in loaded_data['test_log'].value.coords


def test_loads_data_from_non_timeseries_log(load_function: Callable):
    values = np.array([1.1, 2.2, 3.3])
    name = "test_log"
    builder = NexusBuilder()
    builder.add_log(Log(name, values))

    loaded_data = load_function(builder)

    assert np.allclose(loaded_data[name].data.values.values, values)


def test_loads_data_from_multiple_logs_with_same_name(load_function: Callable):
    values_1 = np.array([1.1, 2.2, 3.3])
    values_2 = np.array([4, 5, 6])
    name = "test_log"

    # Add one log to NXentry and the other to an NXdetector,
    # both have the same group name
    builder = NexusBuilder()
    builder.add_log(Log(name, values_1))
    builder.add_detector(Detector(log=Log(name, values_2)))

    with warnings.catch_warnings():
        warnings.filterwarnings("ignore",
                                message='Skipped loading',
                                category=UserWarning)
        loaded_data = load_function(builder)

    assert len(loaded_data) == 2
    for key, da in loaded_data.items():
        assert key.endswith('test_log')
        if 'detector_0' in key:
            assert np.allclose(da.value.values, values_2)
        else:
            assert np.allclose(da.value.values, values_1)


def test_load_instrument_name(load_function: Callable):
    name = "INSTR"
    builder = NexusBuilder()
    builder.add_instrument(name)

    loaded_data = load_function(builder)

    assert loaded_data['instrument_name'].values == name


def test_load_experiment_title(load_function: Callable):
    title = "my experiment"
    builder = NexusBuilder()
    builder.add_title(title)

    loaded_data = load_function(builder)

    assert loaded_data['experiment_title'].values == title


def test_loads_event_and_log_data_from_single_file(load_function: Callable):
    event_time_offsets = np.array([456, 743, 347, 345, 632])
    event_data = EventData(
        event_id=np.array([1, 2, 3, 1, 3]),
        event_time_offset=event_time_offsets,
        event_time_zero=np.array([
            1600766730000000000, 1600766731000000000, 1600766732000000000,
            1600766733000000000
        ]),
        event_index=np.array([0, 3, 3, 5]),
    )

    log_1 = Log("test_log", np.array([1.1, 2.2, 3.3]), np.array([4.4, 5.5, 6.6]))
    log_2 = Log("test_log_2", np.array([123, 253, 756]), np.array([246, 1235, 2369]))

    builder = NexusBuilder()
    builder.add_event_data(event_data)
    builder.add_log(log_1)
    builder.add_log(log_2)

    loaded_data = load_function(builder)

    # Expect time of flight to match the values in the
    # event_time_offset dataset
    # May be reordered due to binning (hence np.sort)
    assert np.allclose(
        np.sort(loaded_data.bins.concat('detector_id').values[0].coords['tof'].values),
        np.sort(event_time_offsets))

    counts_on_detectors = loaded_data.bins.sum()
    # No detector_number dataset in file so expect detector_id to be
    # binned from the min to the max detector_id recorded in event_id
    # dataset: 2 on det 1, 1 on det 2, 2 on det 3
    expected_counts = np.array([[2], [1], [2]])
    assert np.allclose(counts_on_detectors.data.values, expected_counts)
    expected_detector_ids = np.array([1, 2, 3])
    assert np.allclose(loaded_data.coords['detector_id'].values, expected_detector_ids)
    assert "position" not in loaded_data.meta.keys(
    ), "The NXdetectors had no pixel position datasets so we " \
       "should not find 'position' coord"

    # Logs should have been added to the DataArray as attributes
    assert np.allclose(loaded_data.attrs[log_1.name].values.values, log_1.value)
    assert np.allclose(loaded_data.attrs[log_2.name].values.values, log_2.value)


def test_loads_pixel_positions_with_event_data(load_function: Callable):
    pulse_times = np.array([
        1600766730000000000, 1600766731000000000, 1600766732000000000,
        1600766733000000000
    ])
    event_time_offsets_1 = np.array([456, 743, 347, 345, 632])
    event_data_1 = EventData(
        event_id=np.array([1, 2, 3, 1, 3]),
        event_time_offset=event_time_offsets_1,
        event_time_zero=pulse_times,
        event_index=np.array([0, 3, 3, 5]),
    )
    detector_1_ids = np.array([0, 1, 2, 3])
    x_pixel_offset_1 = np.array([0.1, 0.2, 0.1, 0.2])
    y_pixel_offset_1 = np.array([0.1, 0.1, 0.2, 0.2])
    z_pixel_offset_1 = np.array([0.1, 0.2, 0.3, 0.4])

    event_time_offsets_2 = np.array([682, 237, 941, 162, 52])
    event_data_2 = EventData(
        event_id=np.array([4, 5, 6, 4, 6]),
        event_time_offset=event_time_offsets_2,
        event_time_zero=pulse_times,
        event_index=np.array([0, 3, 3, 5]),
    )
    # Multidimensional is fine as long as the shape of
    # the ids and the pixel offsets match
    detector_2_ids = np.array([[4, 5], [6, 7]])
    x_pixel_offset_2 = np.array([[1.1, 1.2], [1.1, 1.2]])
    y_pixel_offset_2 = np.array([[0.1, 0.1], [0.2, 0.2]])

    builder = NexusBuilder()
    offsets_units = "mm"
    builder.add_detector(
        Detector(detector_1_ids,
                 event_data_1,
                 x_offsets=x_pixel_offset_1,
                 y_offsets=y_pixel_offset_1,
                 z_offsets=z_pixel_offset_1,
                 offsets_unit=offsets_units))
    builder.add_detector(
        Detector(detector_2_ids,
                 event_data_2,
                 x_offsets=x_pixel_offset_2,
                 y_offsets=y_pixel_offset_2,
                 offsets_unit=offsets_units))

    loaded_data = load_function(builder)

    # If z offsets are missing they should be zero
    z_pixel_offset_2 = np.array([[0., 0.], [0., 0.]])
    expected_pixel_positions = np.array([
        np.concatenate((x_pixel_offset_1, x_pixel_offset_2.flatten())),
        np.concatenate((y_pixel_offset_1, y_pixel_offset_2.flatten())),
        np.concatenate((z_pixel_offset_1, z_pixel_offset_2.flatten()))
    ]).T
    assert np.allclose(loaded_data.coords['position'].values, expected_pixel_positions)
    assert loaded_data.meta['position'].unit == 'mm'


@pytest.mark.skip(reason="It is unclear if NXdetector with neither data nor event_data "
                  "needs to be considered valid.")
def test_loads_pixel_positions_without_event_data(load_function: Callable):
    """
    This is important in the live-data feature as geometry and event data
    are streamed separately
    """
    detector_1_ids = np.array([0, 1, 2, 3])
    x_pixel_offset_1 = np.array([0.1, 0.2, 0.1, 0.2])
    y_pixel_offset_1 = np.array([0.1, 0.1, 0.2, 0.2])
    z_pixel_offset_1 = np.array([0.1, 0.2, 0.3, 0.4])

    detector_2_ids = np.array([[4, 5], [6, 7]])
    x_pixel_offset_2 = np.array([[1.1, 1.2], [1.1, 1.2]])
    y_pixel_offset_2 = np.array([[0.1, 0.1], [0.2, 0.2]])

    builder = NexusBuilder()
    offsets_units = "mm"
    builder.add_detector(
        Detector(detector_numbers=detector_1_ids,
                 x_offsets=x_pixel_offset_1,
                 y_offsets=y_pixel_offset_1,
                 z_offsets=z_pixel_offset_1,
                 offsets_unit=offsets_units))
    builder.add_detector(
        Detector(detector_numbers=detector_2_ids,
                 x_offsets=x_pixel_offset_2,
                 y_offsets=y_pixel_offset_2,
                 offsets_unit=offsets_units))

    loaded_data = load_function(builder)

    # If z offsets are missing they should be zero
    z_pixel_offset_2 = np.array([[0., 0.], [0., 0.]])
    expected_pixel_positions = np.array([
        np.concatenate((x_pixel_offset_1, x_pixel_offset_2.flatten())),
        np.concatenate((y_pixel_offset_1, y_pixel_offset_2.flatten())),
        np.concatenate((z_pixel_offset_1, z_pixel_offset_2.flatten()))
    ]).T / 1_000  # Divide by 1000 for mm to metres
    assert np.allclose(loaded_data.coords['position'].values, expected_pixel_positions)
    assert loaded_data.meta[
               'base_position'].unit == sc.units.m, \
        "Expected positions to be converted to metres"


def test_loads_pixel_positions_with_no_units(load_function: Callable):
    pulse_times = np.array([
        1600766730000000000, 1600766731000000000, 1600766732000000000,
        1600766733000000000
    ])
    event_time_offsets = np.array([456, 743, 347, 345, 632])
    event_data = EventData(
        event_id=np.array([1, 2, 3, 1, 3]),
        event_time_offset=event_time_offsets,
        event_time_zero=pulse_times,
        event_index=np.array([0, 3, 3, 5]),
    )
    detector_ids = np.array([0, 1, 2, 3])
    x_pixel_offset = np.array([0.1, 0.2, 0.1, 0.2])
    y_pixel_offset = np.array([0.1, 0.1, 0.2, 0.2])
    z_pixel_offset = np.array([0.1, 0.2, 0.3, 0.4])

    builder = NexusBuilder()
    builder.add_detector(
        Detector(detector_ids,
                 event_data,
                 x_offsets=x_pixel_offset,
                 y_offsets=y_pixel_offset,
                 z_offsets=z_pixel_offset,
                 offsets_unit=None))

    loaded_data = load_function(builder)

    assert loaded_data.coords['position'].unit is None


def test_sample_position_at_origin_if_not_explicit_in_file(load_function: Callable):
    # The sample position is the origin of the coordinate
    # system in NeXus files.
    # If there is an NXsample in the file, but it has no "distance" dataset
    # or "depends_on" pointing to NXtransformations then it should be
    # assumed to be at the origin.
    builder = NexusBuilder()
    builder.add_sample(Sample("sample"))
    loaded_data = load_function(builder)

    origin = np.array([0, 0, 0])
    assert np.allclose(loaded_data["sample_position"].values, origin)


def test_loads_multiple_samples(load_function: Callable):
    # More than one sample in the file is possible, but we cannot guess
    # which to record as _the_ "sample_position",
    # instead record them in the form "<GROUP_NAME>_position"
    builder = NexusBuilder()
    sample_1_name = "sample_1"
    sample_2_name = "sample_2"
    builder.add_sample(Sample(sample_1_name))

    distance = 0.762
    units = "m"
    builder.add_sample(Sample(sample_2_name, distance=distance, distance_units=units))
    loaded_data = load_function(builder)

    origin = np.array([0, 0, 0])
    assert np.allclose(
        loaded_data[f"{sample_1_name}_position"].values, origin
    ), "Sample did not have explicit location so expect position " \
       "to be recorded as the origin"
    expected_position = np.array([0, 0, distance])
    assert np.allclose(loaded_data[f"{sample_2_name}_position"].values,
                       expected_position)


def test_skips_loading_source_if_more_than_one_in_file(load_function: Callable):
    # More than one source is a serious error in the file, so load_nexus will not
    # define a source_position rather than guessing which is the "correct" one.
    builder = NexusBuilder()
    builder.add_source(Source("source_1"))
    builder.add_source(Source("source_2"))
    loaded_data = load_function(builder)
    assert 'source_1' in loaded_data
    assert 'source_2' in loaded_data
    assert 'source_position' not in loaded_data


@pytest.mark.parametrize("component_class,component_name",
                         ((Sample, "sample"), (Source, "source")))
def test_component_position_from_distance_dataset_missing_unit(
        component_class: Union[Type[Source], Type[Sample]], component_name: str,
        load_function: Callable):
    builder = NexusBuilder()
    distance = 4.2
    builder.add_component(
        component_class(component_name, distance=distance, distance_units=None))
    loaded_data = load_function(builder)
    assert loaded_data[f'{component_name}_position'].unit is None


@pytest.mark.parametrize("component_class,component_name", [(Sample, "sample"),
                                                            (Source, "source")])
@pytest.mark.parametrize("transform_type,value,value_units,expected_position",
                         ((TransformationType.ROTATION, 0.27, "rad", [0, 0, 0]),
                          (TransformationType.TRANSLATION, 230, "cm", [0, 0, 2.3])))
def test_loads_component_position_from_single_transformation(
        component_class: Union[Type[Source], Type[Sample]], component_name: str,
        transform_type: TransformationType, value: float, value_units: str,
        expected_position: List[float], load_function: Callable):
    builder = NexusBuilder()
    transformation = Transformation(transform_type,
                                    vector=np.array([0, 0, 1]),
                                    value=np.array(value),
                                    value_units=value_units)
    builder.add_component(component_class(component_name, depends_on=transformation))
    loaded_data = load_function(builder)
    assert np.allclose(loaded_data[f"{component_name}_position"].values,
                       expected_position)
    # Resulting position will always be in metres, whatever units are
    # used in the NeXus file
    assert loaded_data[f"{component_name}_position"].unit == sc.Unit("m")


@pytest.mark.parametrize("component_class,component_name", [(Sample, "sample"),
                                                            (Source, "source")])
@pytest.mark.parametrize("transform_type,value,value_units,expected_position",
                         ((TransformationType.ROTATION, 180, "deg", [-1, -2, 3]),
                          (TransformationType.TRANSLATION, 230, "cm", [1, 2, 5.3])))
def test_loads_component_position_from_single_transformation_with_offset(
        component_class: Union[Type[Source], Type[Sample]], component_name: str,
        transform_type: TransformationType, value: float, value_units: str,
        expected_position: List[float], load_function: Callable):
    builder = NexusBuilder()
    transformation = Transformation(transform_type,
                                    vector=np.array([0, 0, 1]),
                                    value=np.array(value),
                                    value_units=value_units,
                                    offset=[1, 2, 3],
                                    offset_unit='m')
    builder.add_component(component_class(component_name, depends_on=transformation))
    loaded_data = load_function(builder)
    assert np.allclose(loaded_data[f"{component_name}_position"].values,
                       expected_position)
    # Resulting position will always be in metres, whatever units are
    # used in the NeXus file
    assert loaded_data[f"{component_name}_position"].unit == sc.Unit("m")


@pytest.mark.parametrize("component_class,component_name", [(Sample, "sample"),
                                                            (Source, "source")])
@pytest.mark.parametrize("transform_type,value,value_units,expected_position",
                         ((TransformationType.ROTATION, 180, "deg", [-1, -2, 3]),
                          (TransformationType.TRANSLATION, 230, "cm", [1, 2, 5.3])))
def test_raises_if_offset_but_not_offset_units_found(
        component_class: Union[Type[Source], Type[Sample]], component_name: str,
        transform_type: TransformationType, value: float, value_units: str,
        expected_position: List[float], load_function: Callable):
    builder = NexusBuilder()
    transformation = Transformation(transform_type,
                                    vector=np.array([0, 0, 1]),
                                    value=np.array(value),
                                    value_units=value_units,
                                    offset=[1, 2, 3],
                                    offset_unit=None)
    builder.add_component(component_class(component_name, depends_on=transformation))
    with warnings.catch_warnings():
        warnings.filterwarnings("ignore",
                                message='Skipped loading',
                                category=UserWarning)
        loaded_data = load_function(builder)
    assert loaded_data is None


@pytest.mark.parametrize("component_class,component_name",
                         ((Sample, "sample"), (Source, "source")))
@pytest.mark.parametrize("transform_type,value,value_units,expected_position",
                         ((TransformationType.ROTATION, 0.27, "rad", [0, 0, 0]),
                          (TransformationType.TRANSLATION, 230, "cm", [0, 0, 2.3])))
def test_loads_component_position_from_log_transformation(
        component_class: Union[Type[Source], Type[Sample]], component_name: str,
        transform_type: TransformationType, value: float, value_units: str,
        expected_position: List[float], load_function: Callable):
    builder = NexusBuilder()
    # Provide "time" data, the builder will write the transformation as
    # an NXlog
    transformation = Transformation(transform_type,
                                    vector=np.array([0, 0, 1]),
                                    value=np.array(value),
                                    value_units=value_units)
    builder.add_component(component_class(component_name, depends_on=transformation))
    loaded_data = load_function(builder)

    # Should load as usual despite the transformation being an NXlog
    # as it only has a single value
    assert np.allclose(loaded_data[f"{component_name}_position"].values,
                       expected_position)
    assert loaded_data[f"{component_name}_position"].unit == sc.Unit("m")


@pytest.mark.parametrize("component_class,component_name",
                         ((Sample, "sample"), (Source, "source")))
@pytest.mark.parametrize(
    "transform_type,value,value_units,expected_position",
    ((TransformationType.ROTATION, [0.27, 0.73], "rad", [0, 0, 0]),
     (TransformationType.TRANSLATION, [230, 310], "cm", [0, 0, 2.3])))
def test_loads_component_position_with_multi_value_log_transformation(
        component_class: Union[Type[Source], Type[Sample]], component_name: str,
        transform_type: TransformationType, value: List[float], value_units: str,
        expected_position: float, load_function: Callable):
    builder = NexusBuilder()
    # Provide "time" data, the builder will write the transformation as
    # an NXlog. This would be encountered in a file from an experiment
    # involving a scan of a motion axis.
    transformation = Transformation(transform_type,
                                    vector=np.array([0, 0, 1]),
                                    value=np.array(value),
                                    time=np.array([1.3, 6.4]),
                                    time_units="s",
                                    value_units=value_units)
    builder.add_component(component_class(component_name, depends_on=transformation))
    loaded_data = load_function(builder)
    assert sc.identical(
        loaded_data[component_name].value['depends_on'].coords["time"],
        sc.Variable(dims=["time"],
                    values=[1300000000, 6400000000],
                    unit="ns",
                    dtype=sc.DType.datetime64))


@pytest.mark.parametrize("component_class,component_name",
                         ((Sample, "sample"), (Source, "source")))
def test_loads_component_position_with_multiple_multi_valued_log_transformations(
        component_class: Union[Type[Source], Type[Sample]], component_name: str,
        load_function: Callable):
    builder = NexusBuilder()
    # Provide "time" data, the builder will write the transformation as
    # an NXlog. This would be encountered in a file from an experiment
    # involving a scan of a motion axis.
    t1 = Transformation(TransformationType.TRANSLATION,
                        vector=np.array([0, 0, 1]),
                        value=np.array([1, 10]),
                        time=np.array([0, 1]),
                        time_units="s",
                        value_units="m")

    t2 = Transformation(TransformationType.TRANSLATION,
                        vector=np.array([0, 0, 1]),
                        value=np.array([5, 50]),
                        time=np.array([0, 1]),
                        time_units="s",
                        value_units="m",
                        depends_on=t1)

    builder.add_component(component_class(component_name, depends_on=t2))
    loaded_data = load_function(builder)
    comp = loaded_data[component_name].value
    assert sc.identical(
        comp['depends_on'].coords["time"],
        sc.Variable(dims=["time"], values=[0, 1], unit="s", dtype=sc.DType.datetime64))


@pytest.mark.parametrize("component_class,component_name",
                         ((Sample, "sample"), (Source, "source")))
def test_multi_valued_log_transformations_time_axis_interpolated_and_trimmed(
        component_class: Union[Type[Source], Type[Sample]], component_name: str,
        load_function: Callable):
    builder = NexusBuilder()
    t1 = Transformation(TransformationType.TRANSLATION,
                        vector=np.array([0, 0, 1]),
                        value=np.array([1, 10]),
                        time=np.array([0, 1]),
                        time_units="s",
                        value_units="m")
    t2 = Transformation(TransformationType.TRANSLATION,
                        vector=np.array([0, 0, 1]),
                        value=np.array([5, 50]),
                        time=np.array([0.1, 1.1]),
                        time_units="s",
                        value_units="m",
                        depends_on=t1)
    builder.add_component(component_class(component_name, depends_on=t2))

    loaded_data = load_function(builder)

    # Note the start at 100 ms, since there is no known value of t2 at 0 ms (when
    # t1 starts)
    comp = loaded_data[component_name].value
    assert sc.identical(
        comp['depends_on'].coords["time"],
        sc.array(dims=["time"],
                 values=[100, 1000, 1100],
                 unit="ms",
                 dtype=sc.DType.datetime64).to(unit="ns"))


@pytest.mark.parametrize("component_class,component_name",
                         ((Sample, "sample"), (Source, "source")))
@pytest.mark.parametrize("transform_type,value_units",
                         ((TransformationType.ROTATION, "deg"),
                          (TransformationType.TRANSLATION, "cm")))
def test_skips_component_position_with_empty_value_log_transformation(
        component_class: Union[Type[Source], Type[Sample]], component_name: str,
        transform_type: TransformationType, value_units: str, load_function: Callable):
    builder = NexusBuilder()
    empty_value = np.array([])
    transformation = Transformation(transform_type,
                                    vector=np.array([0, 0, 1]),
                                    value=empty_value,
                                    time=np.array([1.3, 6.4]),
                                    time_units="s",
                                    value_units=value_units)
    builder.add_component(component_class(component_name, depends_on=transformation))
    with pytest.warns(UserWarning):
        load_function(builder)


@pytest.mark.parametrize("component_class,component_name",
                         ((Sample, "sample"), (Source, "source")))
def test_load_component_position_prefers_transform_over_distance(
        component_class: Union[Type[Source], Type[Sample]], component_name: str,
        load_function: Callable):
    # The "distance" dataset gives the position along the z axis.
    # If there is a "depends_on" pointing to transformations then we
    # prefer to use that instead as it is likely to be more accurate; it
    # can define position and orientation in 3D.
    builder = NexusBuilder()
    transformation = Transformation(TransformationType.TRANSLATION,
                                    vector=np.array([0, 0, 1]),
                                    value=np.array(2.3),
                                    value_units="m")
    builder.add_component(
        component_class(component_name,
                        depends_on=transformation,
                        distance=4.2,
                        distance_units="m"))
    loaded_data = load_function(builder)

    expected_position = np.array([0, 0, transformation.value])
    assert np.allclose(loaded_data[f"{component_name}_position"].values,
                       expected_position)
    assert loaded_data[f"{component_name}_position"].unit == sc.Unit("m")


@pytest.mark.parametrize("component_class,component_name",
                         ((Sample, "sample"), (Source, "source")))
@pytest.mark.parametrize("transform_type",
                         (TransformationType.ROTATION, TransformationType.TRANSLATION))
def test_skips_component_position_from_transformation_missing_unit(
        component_class: Union[Type[Source], Type[Sample]], component_name: str,
        transform_type: TransformationType, load_function: Callable):
    builder = NexusBuilder()
    transformation = Transformation(transform_type,
                                    np.array([0, 0, -1]),
                                    value=np.array(2.3))
    builder.add_component(component_class(component_name, depends_on=transformation))
    with pytest.warns(UserWarning):
        load_function(builder)


@pytest.mark.parametrize("component_class,component_name",
                         ((Sample, "sample"), (Source, "source")))
def test_loads_component_position_from_multiple_transformations(
        component_class: Union[Type[Source], Type[Sample]], component_name: str,
        load_function: Callable):
    builder = NexusBuilder()
    transformation_1 = Transformation(TransformationType.ROTATION,
                                      np.array([0, 1, 0]),
                                      value=np.array(90),
                                      value_units="deg")
    transformation_2 = Transformation(TransformationType.TRANSLATION,
                                      np.array([0, 0, 1]),
                                      value=np.array(2.3),
                                      value_units="m",
                                      depends_on=transformation_1)
    builder.add_component(component_class(component_name, depends_on=transformation_2))
    loaded_data = load_function(builder)

    expected_position = np.array([2.3, 0, 0])
    assert np.allclose(loaded_data[f"{component_name}_position"].values,
                       expected_position)
    assert loaded_data[f"{component_name}_position"].unit == sc.Unit("m")


def test_skips_source_position_if_not_given_in_file(load_function: Callable):
    builder = NexusBuilder()
    builder.add_source(Source("source"))
    loaded_data = load_function(builder)
    assert 'source_position' not in loaded_data


@pytest.mark.parametrize("component_class,component_name",
                         ((Sample, "sample"), (Source, "source")))
def test_loads_component_position_from_distance_dataset(
        component_class: Union[Type[Source], Type[Sample]], component_name: str,
        load_function: Callable):
    # If the NXsource or NXsample contains a "distance" dataset
    # this gives the position along the z axis. If there was a "depends_on"
    # pointing to transformations then we'd use that instead as it is
    # likely to be more accurate; it can define position and orientation in 3D.
    builder = NexusBuilder()
    distance = 4.2
    units = "m"
    builder.add_component(
        component_class(component_name, distance=distance, distance_units=units))
    loaded_data = load_function(builder)

    expected_position = np.array([0, 0, distance])
    assert np.allclose(loaded_data[f"{component_name}_position"].values,
                       expected_position)
    assert loaded_data[f"{component_name}_position"].unit == sc.Unit(units)


def test_loads_source_position_dependent_on_sample_position(load_function: Callable):
    builder = NexusBuilder()
    transformation_0 = Transformation(TransformationType.ROTATION,
                                      np.array([0, 1, 0]),
                                      value=np.array(90),
                                      value_units="deg")
    transformation_1 = Transformation(TransformationType.TRANSLATION,
                                      np.array([0, 0, 1]),
                                      value=np.array(2.3),
                                      value_units="m",
                                      depends_on=transformation_0)
    builder.add_sample(Sample("sample", depends_on=transformation_1))
    transformation_2 = Transformation(
        TransformationType.TRANSLATION,
        np.array([0, 0, 1]),
        value=np.array(1.0),
        value_units="m",
        depends_on="/entry/sample/transformations/transform_1")
    builder.add_source(Source("source", depends_on=transformation_2))
    loaded_data = load_function(builder)

    expected_position = np.array([3.3, 0, 0])
    assert np.allclose(loaded_data["source_position"].values, expected_position)
    assert loaded_data["source_position"].unit == sc.Unit("m")


def test_loads_pixel_positions_with_transformations(load_function: Callable):
    pulse_times = np.array([
        1600766730000000000, 1600766731000000000, 1600766732000000000,
        1600766733000000000
    ])
    event_time_offsets_1 = np.array([456, 743, 347, 345, 632])
    event_data_1 = EventData(
        event_id=np.array([1, 2, 3, 1, 3]),
        event_time_offset=event_time_offsets_1,
        event_time_zero=pulse_times,
        event_index=np.array([0, 3, 3, 5]),
    )
    detector_1_ids = np.array([0, 1, 2, 3])
    x_pixel_offset_1 = np.array([0.1, 0.2, 0.1, 0.2])
    y_pixel_offset_1 = np.array([0.1, 0.1, 0.2, 0.2])
    z_pixel_offset_1 = np.array([0.1, 0.2, 0.3, 0.4])

    distance = 57  # cm
    transformation = Transformation(TransformationType.TRANSLATION,
                                    vector=np.array([0, 0, 1]),
                                    value=np.array(distance),
                                    value_units="cm")

    builder = NexusBuilder()
    builder.add_detector(
        Detector(detector_1_ids,
                 event_data_1,
                 x_offsets=x_pixel_offset_1,
                 y_offsets=y_pixel_offset_1,
                 z_offsets=z_pixel_offset_1,
                 offsets_unit="m",
                 depends_on=transformation))

    loaded_data = load_function(builder)

    expected_pixel_positions = np.array(
        [x_pixel_offset_1, y_pixel_offset_1, z_pixel_offset_1]).T
    assert np.allclose(loaded_data.meta['base_position'].values,
                       expected_pixel_positions)

    expected_transform = sc.spatial.affine_transform(unit=sc.units.m,
                                                     value=[[1, 0, 0, 0], [0, 1, 0, 0],
                                                            [0, 0, 1, 0.57],
                                                            [0, 0, 0, 1]])

    assert np.allclose(loaded_data.meta['position_transformations'].value.values,
                       expected_transform.values)

    assert np.allclose(
        (loaded_data.meta["position_transformations"].value *
         loaded_data.meta["base_position"]["detector_id", 0]).values,
        [0.1, 0.1, 0.67],
    )

    assert np.allclose(loaded_data.coords["position"]["detector_id", 0].values,
                       [0.1, 0.1, 0.67])


def test_loads_pixel_positions_with_multiple_transformations(load_function: Callable):
    event_data_1 = EventData(
        event_id=np.array([0, 0, 0, 0, 0]),
        event_time_offset=(np.array([456, 743, 347, 345, 632])),
        event_time_zero=np.array([
            1600766730000000000, 1600766731000000000, 1600766732000000000,
            1600766733000000000
        ]),
        event_index=np.array([0, 3, 3, 5]),
    )
    event_data_2 = EventData(
        event_id=np.array([1, 1, 1, 1, 1]),
        event_time_offset=(np.array([456, 743, 347, 345, 632])),
        event_time_zero=np.array([
            1600766730000000000, 1600766731000000000, 1600766732000000000,
            1600766733000000000
        ]),
        event_index=np.array([0, 3, 3, 5]),
    )

    transformation1 = Transformation(TransformationType.TRANSLATION,
                                     vector=np.array([0, 0, 1]),
                                     value=np.array(12),
                                     value_units="cm")
    transformation2 = Transformation(TransformationType.TRANSLATION,
                                     vector=np.array([0, 0, 1]),
                                     value=np.array(34),
                                     value_units="cm")

    builder = NexusBuilder()
    builder.add_detector(
        Detector(np.array([0]),
                 event_data_1,
                 x_offsets=np.array([0.1]),
                 y_offsets=np.array([0.1]),
                 z_offsets=np.array([0.1]),
                 offsets_unit="m",
                 depends_on=transformation1))

    builder.add_detector(
        Detector(np.array([1]),
                 event_data_2,
                 x_offsets=np.array([0.6]),
                 y_offsets=np.array([0.6]),
                 z_offsets=np.array([0.6]),
                 offsets_unit="m",
                 depends_on=transformation2))

    loaded_data = load_function(builder)

    assert np.allclose(
        (loaded_data.meta["position_transformations"]["detector_id", 0].value *
         loaded_data.meta["base_position"]["detector_id", 0]).values,
        [0.1, 0.1, 0.1 + 0.12],
    )
    assert np.allclose(
        (loaded_data.meta["position_transformations"]["detector_id", 1].value *
         loaded_data.meta["base_position"]["detector_id", 1]).values,
        [0.6, 0.6, 0.6 + 0.34],
    )


def test_links_to_event_data_group_are_ignored(load_function: Callable):
    event_time_offsets = np.array([456, 743, 347, 345, 632])
    event_data = EventData(
        event_id=np.array([1, 2, 3, 1, 3]),
        event_time_offset=event_time_offsets,
        event_time_zero=np.array([
            1600766730000000000, 1600766731000000000, 1600766732000000000,
            1600766733000000000
        ]),
        event_index=np.array([0, 3, 3, 5]),
    )

    builder = NexusBuilder()
    builder.add_event_data(event_data)
    builder.add_hard_link(Link("/entry/hard_link_to_events", "/entry/events_0"))
    builder.add_soft_link(Link("/entry/soft_link_to_events", "/entry/events_0"))

    loaded_data = load_function(builder)

    # The output Variable must contain the events from the added event
    # dataset with no duplicate data due to the links

    # Expect time of flight to match the values in the
    # event_time_offset dataset
    # May be reordered due to binning (hence np.sort)
    assert np.array_equal(
        np.sort(loaded_data.bins.concat('detector_id').values[0].coords['tof'].values),
        np.sort(event_time_offsets))

    counts_on_detectors = loaded_data.bins.sum()
    # No detector_number dataset in file so expect detector_id to be
    # binned according to whatever detector_ids are present in event_id
    # dataset: 2 on det 1, 1 on det 2, 2 on det 3
    expected_counts = np.array([[2], [1], [2]])
    assert np.array_equal(counts_on_detectors.data.values, expected_counts)
    expected_detector_ids = np.array([1, 2, 3])
    assert np.array_equal(loaded_data.coords['detector_id'].values,
                          expected_detector_ids)


def test_links_in_transformation_paths_are_followed(load_function: Callable):
    builder = NexusBuilder()
    distance = 13.6
    builder.add_component(Source("source"))
    builder.add_dataset_at_path(
        "/entry/transform", np.array(distance), {
            "vector": np.array([0, 0, 1]),
            "units": "m",
            "transformation_type": "translation",
            "depends_on": "."
        })
    builder.add_dataset_at_path("/entry/source/depends_on", "/entry/transform_link", {})
    builder.add_soft_link(Link("/entry/transform_link", "/entry/transform"))
    loaded_data = load_function(builder)

    assert np.allclose(loaded_data["source_position"].values, [0, 0, distance])
    # Resulting position will always be in metres, whatever units are
    # used in the NeXus file
    assert loaded_data["source_position"].unit == sc.Unit("m")


def test_relative_links_in_transformation_paths_are_followed(load_function: Callable):
    builder = NexusBuilder()
    distance = 13.6
    builder.add_component(Source("source"))
    builder.add_dataset_at_path(
        "/entry/transform1", np.array(distance), {
            "vector": np.array([0, 0, 1]),
            "units": "m",
            "transformation_type": "translation",
            "depends_on": "."
        })
    builder.add_dataset_at_path(
        "/entry/transform2", np.array(distance), {
            "vector": np.array([0, 0, 1]),
            "units": "m",
            "transformation_type": "translation",
            "depends_on": "transform1"
        })
    builder.add_dataset_at_path("/entry/source/depends_on", "/entry/transform2", {})
    loaded_data = load_function(builder)

    assert np.allclose(loaded_data["source_position"].values, [0, 0, 2 * distance])
    # Resulting position will always be in metres, whatever units are
    # used in the NeXus file
    assert loaded_data["source_position"].unit == sc.Unit("m")


def test_linked_datasets_are_found(load_function: Callable):
    event_data = EventData(
        event_id=np.array([1, 2, 3, 1, 3]),
        event_time_offset=np.array([456, 743, 347, 345, 632]),
        event_time_zero=np.array([
            1600766730000000000, 1600766731000000000, 1600766732000000000,
            1600766733000000000
        ]),
        event_index=np.array([0, 3, 3, 5]),
    )

    builder = NexusBuilder()
    builder.add_event_data(event_data)
    replaced_ids = np.array([1, 1, 1, 2, 3])
    builder.add_dataset_at_path("/entry/ids", replaced_ids, {})
    replaced_tofs = np.array([273, 546, 573, 812, 932])
    builder.add_dataset_at_path("/entry/tofs", replaced_tofs, {})
    # Replace dataset in the NXevent_data with a link to the
    # replacement dataset
    builder.add_hard_link(Link("/entry/events_0/event_id", "/entry/ids"))
    builder.add_soft_link(Link("/entry/events_0/event_time_offset", "/entry/tofs"))

    loaded_data = load_function(builder)

    assert np.array_equal(
        np.sort(loaded_data.bins.concat('detector_id').values[0].coords['tof'].values),
        np.sort(replaced_tofs))

    counts_on_detectors = loaded_data.bins.sum()
    expected_counts = np.array([[3], [1], [1]])
    assert np.array_equal(counts_on_detectors.data.values, expected_counts)
    expected_detector_ids = np.array([1, 2, 3])
    assert np.array_equal(loaded_data.coords['detector_id'].values,
                          expected_detector_ids)


def test_loads_sample_ub_matrix(load_function: Callable):
    builder = NexusBuilder()
    builder.add_component(Sample("sample", ub_matrix=np.ones(shape=[3, 3])))
    loaded_data = load_function(builder)
    sample = loaded_data['sample'].value
    assert sc.identical(
        sample['ub_matrix'],
        sc.spatial.linear_transform(value=np.ones(shape=[3, 3]),
                                    unit=sc.units.angstrom**-1))
    assert "orientation_matrix" not in sample


def test_loads_sample_orientation_matrix(load_function: Callable):
    builder = NexusBuilder()
    builder.add_component(Sample("sample", orientation_matrix=np.ones(shape=[3, 3])))
    loaded_data = load_function(builder)
    sample = loaded_data['sample'].value
    assert sc.identical(
        sample["orientation_matrix"],
        sc.spatial.linear_transform(value=np.ones(shape=[3, 3]), unit=sc.units.one))
    assert "ub_matrix" not in sample


def test_loads_multiple_sample_ub_matrix(load_function: Callable):
    builder = NexusBuilder()
    builder.add_component(Sample("sample1", ub_matrix=np.ones(shape=[3, 3])))
    builder.add_component(Sample("sample2", ub_matrix=np.identity(3)))
    builder.add_component(Sample("sample3"))  # No ub specified
    loaded_data = load_function(builder)
    assert sc.identical(
        loaded_data["sample1"].value["ub_matrix"],
        sc.spatial.linear_transform(value=np.ones(shape=[3, 3]),
                                    unit=sc.units.angstrom**-1))
    assert sc.identical(
        loaded_data["sample2"].value["ub_matrix"],
        sc.spatial.linear_transform(value=np.identity(3), unit=sc.units.angstrom**-1))
    assert "ub_matrix" not in loaded_data['sample3'].value


def test_warning_but_no_error_for_unrecognised_log_unit(load_function: Callable):
    values = np.array([1.1, 2.2, 3.3])
    times = np.array([4.4, 5.5, 6.6])
    name = "test_log"
    builder = NexusBuilder()
    unknown_unit = "elephants"
    builder.add_log(Log(name, values, times, value_units=unknown_unit, time_units="s"))

    with pytest.warns(UserWarning):
        loaded_data = load_function(builder)

    # Expect a sc.Dataset with log names as keys
    assert name in loaded_data


def test_start_and_end_times_appear_in_dataset_if_set(load_function: Callable):
    builder = NexusBuilder()
    builder.add_run_start_time("2001-01-01T00:00:00")
    builder.add_run_end_time("2002-02-02T00:00:00")

    loaded_data = load_function(builder)

    assert sc.identical(loaded_data["start_time"],
                        sc.DataArray(sc.scalar("2001-01-01T00:00:00")))
    assert sc.identical(loaded_data["end_time"],
                        sc.DataArray(sc.scalar("2002-02-02T00:00:00")))


@pytest.mark.parametrize("log_start,scaling_factor",
                         (("2000-01-01T01:00:00", 1000), ("2000-01-01T00:00:00", 0.001),
                          ("2010-01-01T00:00:00", None)))
def test_load_log_times(log_start: str, scaling_factor: float, load_function: Callable):

    times = np.array([0., 10., 20., 30., 40., 50.], dtype="float64")

    builder = NexusBuilder()
    builder.add_log(
        Log(name="test_log",
            value=np.zeros(shape=(len(times), )),
            time=times,
            start_time=log_start,
            scaling_factor=scaling_factor))

    loaded_data = load_function(builder)

    times_ns = sc.to_unit(
        sc.array(dims=["time"],
                 values=times * (scaling_factor or 1.0),
                 unit=sc.units.s,
                 dtype=sc.DType.float64), sc.units.ns).astype(sc.DType.int64)

    expected = sc.scalar(value=np.datetime64(log_start),
                         unit=sc.units.ns,
                         dtype=sc.DType.datetime64) + times_ns

    actual = loaded_data["test_log"].values.coords['time']

    diffs = np.abs(actual.values - expected.values)

    # Allow 1ns difference for rounding errors between different routes
    assert all(diffs <= np.array(1).astype("timedelta64[ns]"))


def test_load_log_times_when_logs_do_not_have_start_time(load_function: Callable):
    # If an NXLog doesn't have a start time attribute then 1970-01-01 should be
    # assumed instead

    times = np.array([-10., 0., 10., 20., 30., 40., 50.], dtype="float64")

    builder = NexusBuilder()
    builder.add_log(
        Log(name="test_log",
            value=np.zeros(shape=(len(times), )),
            time=times,
            start_time=None))

    loaded_data = load_function(builder)

    times_ns = sc.to_unit(
        sc.array(dims=["time"], values=times, unit=sc.units.s, dtype=sc.DType.float64),
        sc.units.ns).astype(sc.DType.int64)

    expected = sc.scalar(value=np.datetime64("1970-01-01T00:00:00"),
                         unit=sc.units.ns,
                         dtype=sc.DType.datetime64) + times_ns

    actual = loaded_data["test_log"].values.coords['time']

    diffs = np.abs(actual.values - expected.values)

    # Allow 1ns difference for rounding errors between different routes
    assert all(diffs <= np.array(1).astype("timedelta64[ns]"))


@pytest.mark.parametrize("units", ("ps", "ns", "us", "ms", "s", "minute", "hour"))
def test_adjust_log_times_with_different_time_units(units, load_function: Callable):

    times = [1, 2, 3]

    builder = NexusBuilder()
    builder.add_log(
        Log(name="test_log",
            value=np.zeros(shape=(len(times), )),
            time=np.array(times, dtype="float64"),
            time_units=units,
            start_time="2000-01-01T00:00:00"), )

    loaded_data = load_function(builder)

    times_ns = sc.to_unit(
        sc.array(dims=["time"], values=times, unit=units, dtype=sc.DType.int64),
        sc.units.ns)

    expected = sc.scalar(value=np.datetime64("2000-01-01T00:00:00"),
                         unit=sc.units.ns,
                         dtype=sc.DType.datetime64) + times_ns

    actual = loaded_data["test_log"].values.coords['time']

    diffs = np.abs(actual.values - expected.values)

    # Allow 1ns difference for rounding errors between different routes
    assert all(diffs <= np.array(1).astype("timedelta64[ns]"))


def test_nexus_file_with_invalid_nxlog_time_units_loads_dataset_as_non_datetime(
        load_function: Callable):
    builder = NexusBuilder()
    builder.add_log(
        Log(
            name="test_log_1",
            value=np.zeros(shape=(1, )),
            time=np.array([1.0]),
            time_units="m",  # Time in metres, should fail.
            start_time="1970-01-01T00:00:00"))

    loaded_data = load_function(builder)
    assert loaded_data['test_log_1'].value.coords['time'].dtype == sc.DType('float64')
    assert loaded_data['test_log_1'].value.coords['time'].unit == 'm'


def test_nexus_file_with_invalid_log_start_date_uses_epoch(load_function: Callable):
    builder = NexusBuilder()
    builder.add_log(
        Log(name="test_log_1",
            value=np.zeros(shape=(1, )),
            time=np.array([1]),
            start_time="this_isnt_a_valid_log_start_time"))
    builder.add_log(
        Log(name="test_log_2",
            value=np.zeros(shape=(1, )),
            time=np.array([1]),
            start_time="1970-01-01T00:00:00"))

    loaded_data = load_function(builder)
    assert sc.identical(
        loaded_data['test_log_1'].value.coords['time'],
        sc.epoch(unit='s') +
        sc.array(dims=['time'], values=[1], unit='s', dtype='int64'))


def test_extended_ascii_in_ascii_encoded_dataset(load_function: Callable):
    if load_function == load_from_json:
        pytest.skip("JSON serialiser can only serialize strings, not bytes.")

    builder = NexusBuilder()
    # When writing, if we use bytes h5py will write as ascii encoding
    # 0xb0 = degrees symbol in latin-1 encoding.
    builder.add_title(b"run at rot=90" + bytes([0xb0]))

    with pytest.warns(UserWarning, match="contains characters in extended ascii range"):
        loaded_data = load_function(builder)

        assert sc.identical(loaded_data["experiment_title"],
                            sc.DataArray(data=sc.scalar("run at rot=90°")))


@pytest.mark.parametrize("test_string", UTF8_TEST_STRINGS)
def test_utf8_encoded_dataset(load_function: Callable, test_string):
    builder = NexusBuilder()
    # When writing, if we use str h5py will write as utf8 encoding
    builder.add_title(test_string)

    loaded_data = load_function(builder)

    assert sc.identical(loaded_data["experiment_title"],
                        sc.DataArray(data=sc.scalar(test_string)))


def test_extended_ascii_in_ascii_encoded_attribute(load_function: Callable):
    if load_function == load_from_json:
        pytest.skip("JSON serialiser can only serialize strings, not bytes.")

    builder = NexusBuilder()
    # When writing, if we use bytes h5py will write as ascii encoding
    # 0xb0 = degrees symbol in latin-1 encoding.
    builder.add_log(Log(name="testlog", value_units=bytes([0xb0]), value=np.array([0])))

    with pytest.warns(UserWarning, match="contains characters in extended ascii range"):
        loaded_data = load_function(builder)

        assert loaded_data["testlog"].data.values.unit == sc.units.deg


# Can't use UTF-8 test strings as above for this test as the units need to be valid.
# Just do a single test with degrees.
def test_utf8_encoded_attribute(load_function: Callable):
    builder = NexusBuilder()
    # When writing, if we use str h5py will write as utf8 encoding
    builder.add_log(Log(name="testlog", value_units="°", value=np.array([0])))

    loaded_data = load_function(builder)
    assert loaded_data["testlog"].data.values.unit == sc.units.deg


def test_load_nexus_adds_single_tof_bin(load_function: Callable):
    event_time_offsets = np.array([456, 743, 347, 345, 632], dtype="float64")
    event_data = EventData(
        event_id=np.array([1, 2, 3, 1, 3]),
        event_time_offset=event_time_offsets,
        event_time_zero=np.array([
            1600766730000000000, 1600766731000000000, 1600766732000000000,
            1600766733000000000
        ]),
        event_index=np.array([0, 3, 3, 5]),
    )

    builder = NexusBuilder()
    builder.add_event_data(event_data)

    loaded_data = load_function(builder)

    # Size 2 for each of the two bin edges around a single bin
    assert loaded_data.coords["tof"].shape == (2, )

    # Assert bin edges correspond to smallest and largest+1 time-of-flights
    # in data.
    assert sc.identical(loaded_data.coords["tof"]["tof", 0],
                        sc.scalar(value=np.min(event_time_offsets), unit=sc.units.ns))
    assert sc.identical(
        loaded_data.coords["tof"]["tof", 1],
        sc.scalar(value=np.nextafter(np.max(event_time_offsets), float("inf")),
                  unit=sc.units.ns))


def test_nexus_file_with_choppers(load_function: Callable):
    builder = NexusBuilder()
    builder.add_instrument("dummy")
    builder.add_chopper(
        Chopper("chopper_1",
                distance=10.0,
                rotation_speed=60.0,
                rotation_units="Hz",
                distance_units="m"))
    loaded_data = load_function(builder)
    chopper = loaded_data['chopper_1'].value
    assert sc.identical(chopper["rotation_speed"], 60.0 * sc.Unit("Hz"))
    assert sc.identical(chopper["distance"], 10.0 * sc.Unit("m"))


def test_nexus_file_with_two_choppers(load_function: Callable):
    builder = NexusBuilder()
    builder.add_instrument("dummy")
    builder.add_chopper(
        Chopper("chopper_1",
                distance=11.0 * 1000,
                rotation_speed=65.0 / 1000,
                rotation_units="MHz",
                distance_units="mm"))
    builder.add_chopper(
        Chopper("chopper_2",
                distance=10.0,
                rotation_speed=60.0,
                rotation_units="Hz",
                distance_units="m"))
    loaded_data = load_function(builder)
    chopper1 = loaded_data['chopper_1'].value
    chopper2 = loaded_data['chopper_2'].value
    assert sc.identical(chopper1["rotation_speed"], (65.0 / 1000) * sc.Unit("MHz"))
    assert sc.identical(chopper1["distance"], (11.0 * 1000) * sc.Unit("mm"))
    assert sc.identical(chopper2["rotation_speed"], 60.0 * sc.Unit("Hz"))
    assert sc.identical(chopper2["distance"], 10.0 * sc.Unit("m"))


def test_load_monitors(load_function: Callable):
    builder = NexusBuilder()

    # Monitor with data
    builder.add_monitor(
        Monitor(
            name="monitor1",
            data=np.ones(shape=(2, 4, 6), dtype="float64"),
            axes=[
                ("event_index", np.arange(3, 5, dtype="float64")),
                ("period_index", np.arange(3, 7, dtype="float64")),
                ("time_of_flight", np.arange(3, 9, dtype="float64")),
            ],
        ))

    # Monitor with only one "axis" and only one data item
    builder.add_monitor(
        Monitor("monitor2",
                data=np.array([1.]),
                axes=[("time_of_flight", np.array([1.]))]))

    assert sc.identical(
        load_function(builder)["monitor1"].data.values,
        sc.DataArray(data=sc.ones(
            dims=["event_index", "period_index", "time_of_flight"],
            shape=(2, 4, 6),
            dtype=sc.DType.float64),
                     coords={
                         "event_index":
                         sc.Variable(dims=["event_index"],
                                     values=np.arange(3, 5, dtype="float64")),
                         "period_index":
                         sc.Variable(dims=["period_index"],
                                     values=np.arange(3, 7, dtype="float64")),
                         "time_of_flight":
                         sc.Variable(dims=["time_of_flight"],
                                     values=np.arange(3, 9, dtype="float64")),
                     }))

    assert sc.identical(
        load_function(builder)["monitor2"].data.values,
        sc.DataArray(data=sc.ones(dims=["time_of_flight"], shape=(1, )),
                     coords={
                         "time_of_flight":
                         sc.Variable(dims=["time_of_flight"], values=np.array([1.])),
                     }))


def test_load_monitors_with_event_mode_data(load_function: Callable):
    builder = NexusBuilder()

    # Monitor with data
    builder.add_monitor(
        Monitor(name="monitor1",
                data=np.ones(shape=(2, 3, 4), dtype="float64"),
                axes=[
                    ("event_index", np.array([0, 5], dtype="int64")),
                    ("period_index", np.array([2, 3, 4], dtype="int64")),
                    ("time_of_flight", np.array([3, 4, 5, 6], dtype="float64")),
                ],
                events=EventData(
                    event_id=np.array([0, 0, 0, 0, 0], dtype="int64"),
                    event_time_offset=np.array([1, 2, 3, 4, 5], dtype="float64"),
                    event_time_zero=np.array([0, 1], dtype="float64"),
                    event_index=np.array([0, 5], dtype="int64"),
                )))

    # Monitor with only one "axis" and only one data item
    builder.add_monitor(
        Monitor("monitor2",
                data=np.array([1.], dtype="float64"),
                axes=[("time_of_flight", np.array([1.], dtype="float64"))],
                events=EventData(
                    event_id=np.array([1, 1, 1, 1, 1], dtype="int64"),
                    event_time_offset=np.array([6, 7, 8, 9, 10], dtype="float64"),
                    event_time_zero=np.array([0, 1], dtype="float64"),
                    event_index=np.array([0, 5], dtype="int64"),
                )))

    with warnings.catch_warnings():
        warnings.filterwarnings("ignore", message='Skipped load', category=UserWarning)
        mon_1_events = load_function(builder)["monitor1"].data.values
        mon_2_events = load_function(builder)["monitor2"].data.values

    assert sc.identical(
        mon_1_events.values[0].coords["tof"],
        sc.array(dims=["event"],
                 values=[1, 2, 3, 4, 5],
                 unit=sc.units.ns,
                 dtype=sc.DType.float64))
    assert sc.identical(
        mon_2_events.values[0].coords["tof"],
        sc.array(dims=["event"],
                 values=[6, 7, 8, 9, 10],
                 unit=sc.units.ns,
                 dtype=sc.DType.float64))


def test_load_monitor_with_transformation(load_function: Callable):
    builder = NexusBuilder()

    transformation = Transformation(TransformationType.TRANSLATION,
                                    vector=np.array([0, 0, 1]),
                                    value=np.array(6.5),
                                    value_units='m',
                                    offset=[1, 2, 3],
                                    offset_unit='m')

    # Monitor with data and transformation
    builder.add_monitor(
        Monitor(name="monitor1",
                data=np.ones(shape=(2, 4, 6), dtype="float64"),
                axes=[
                    ("event_index", np.arange(3, 5, dtype="float64")),
                    ("period_index", np.arange(3, 7, dtype="float64")),
                    ("time_of_flight", np.arange(3, 9, dtype="float64")),
                ],
                depends_on=transformation))

    loaded = load_function(builder)["monitor1"].data.values

    assert sc.identical(loaded.coords["position"],
                        sc.vector(value=[1., 2., 9.5], unit="m"))


def test_load_nexus_file_containing_empty_arrays(load_function: Callable):
    event_data = EventData(
        event_id=np.array([], dtype='int32'),
        event_time_offset=np.array([]),
        event_time_zero=np.array([]),
        event_index=np.array([]),
    )

    builder = NexusBuilder()
    builder.add_event_data(event_data)
    builder.add_log(Log("test_log", np.array([0, 1, 2]), np.array([4, 5, 6])))

    # Empty datasets should not stop other data (e.g. metadata) from being loaded.
    loaded_data = load_function(builder)
    assert "test_log" in loaded_data.attrs
    assert all(loaded_data.attrs["test_log"].value.values == np.array([0, 1, 2]))



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2022 Scipp contributors (https://github.com/scipp)
# @author Simon Heybrock

# flake8: noqa

import os

from .conversions import convert, conversion_graph, deduce_conversion_graph



import scippneutron as scn
import scipp as sc
import numpy as np
from .common import make_dataset_with_beamline
import pytest
import plopp as pp

import matplotlib

matplotlib.use('Agg')
pp.patch_scipp()


def test_neutron_instrument_view_3d():
    d = make_dataset_with_beamline()
    scn.instrument_view(d["a"])


def test_neutron_instrument_view_with_masks():
    d = make_dataset_with_beamline()
    x = np.transpose(d.coords['position'].values)[0, :]
    d['a'].masks['amask'] = sc.Variable(dims=['position'],
                                        values=np.less(np.abs(x), 0.5))
    scn.instrument_view(d["a"])


def test_neutron_instrument_view_with_cmap_args():
    d = make_dataset_with_beamline()
    scn.instrument_view(d["a"],
                        vmin=0.001 * sc.units.one,
                        vmax=5.0 * sc.units.one,
                        cmap="magma",
                        norm="log")


def _make_component_settings(*,
                             data,
                             center='sample_position',
                             type='box',
                             size_unit=sc.units.m,
                             wireframe=False,
                             component_size=(0.1, 0.1, 0.1)):
    comp_size = sc.vector(value=component_size, unit=size_unit) if isinstance(
        component_size, tuple) else sc.scalar(value=component_size, unit=size_unit)
    center = data.meta[center] if isinstance(center, str) else center
    sample_settings = {'center': center, 'size': comp_size, 'type': type}
    return sample_settings


def test_neutron_instrument_view_components_valid():
    d = make_dataset_with_beamline()
    scn.instrument_view(d["a"], components={'sample': _make_component_settings(data=d)})


def test_neutron_instrument_view_with_size_scalar():
    d = make_dataset_with_beamline()
    scn.instrument_view(
        d["a"],
        components={'sample': _make_component_settings(data=d, component_size=2.0)})


def test_neutron_instrument_view_components_multiple_valid():
    d = make_dataset_with_beamline()
    scn.instrument_view(d["a"],
                        components={
                            'sample':
                            _make_component_settings(data=d, center='sample_position'),
                            'source':
                            _make_component_settings(data=d, center='source_position')
                        })


def test_neutron_instrument_view_components_with_non_beamline_component():
    d = make_dataset_with_beamline()
    widget_center = sc.vector(value=[1, 1, 1], unit=sc.units.m)
    scn.instrument_view(
        d["a"],
        components={'widget': _make_component_settings(data=d, center=widget_center)})


def test_neutron_instrument_view_components_with_wireframe():
    d = make_dataset_with_beamline()
    scn.instrument_view(
        d["a"],
        components={'sample': _make_component_settings(data=d, wireframe=False)})
    scn.instrument_view(
        d["a"], components={'sample': _make_component_settings(data=d, wireframe=True)})


def test_neutron_instrument_view_components_with_invalid_type():
    d = make_dataset_with_beamline()
    # Check that all our valid shape types work
    for shape_type in ['box', 'cylinder', 'disk']:
        scn.instrument_view(
            d["a"],
            components={'sample': _make_component_settings(data=d, type=shape_type)})
    with pytest.raises(ValueError):
        scn.instrument_view(d["a"],
                            components={
                                'sample':
                                _make_component_settings(data=d, type='trefoil_knot')
                            })


def test_neutron_instrument_view_components_with_invalid_size_unit():
    d = make_dataset_with_beamline()
    # mm is fine, source_position is set in meters. Just a scaling factor.
    scn.instrument_view(
        d["a"],
        components={'sample': _make_component_settings(data=d, size_unit=sc.units.mm)})
    # cannot scale us to meters. This should throw
    with pytest.raises(sc.core.UnitError):
        scn.instrument_view(d["a"],
                            components={
                                'sample':
                                _make_component_settings(data=d, size_unit=sc.units.us)
                            })



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2022 Scipp contributors (https://github.com/scipp)
# @author Simon Heybrock
"""
Frame transformations for time-of-flight neutron scattering data.
"""
import scipp as sc
from scipp.constants import h, m_n
from ..conversion.graph.beamline import Ltotal
from .._utils import as_float_type, elem_unit


def _tof_from_wavelength(*, wavelength, Ltotal):
    scale = (m_n / h).to(unit=sc.units.us / elem_unit(Ltotal) / elem_unit(wavelength))
    return as_float_type(Ltotal * scale, wavelength) * wavelength


def _tof_to_time_offset(*, tof, frame_length, frame_offset):
    unit = tof.unit
    frame_length = frame_length.to(unit=unit)
    arrival_time_offset = frame_offset.to(unit=unit) + tof
    time_offset = arrival_time_offset % frame_length
    return time_offset


def _time_offset_to_tof(*, time_offset, time_offset_pivot, tof_min, frame_length):
    frame_length = frame_length.to(unit=elem_unit(time_offset))
    time_offset_pivot = time_offset_pivot.to(unit=elem_unit(time_offset))
    tof_min = tof_min.to(unit=elem_unit(time_offset))
    shift = tof_min - time_offset_pivot
    tof = sc.where(time_offset >= time_offset_pivot, shift, shift + frame_length)
    tof += time_offset
    return tof


def to_tof():
    """
    Return a graph suitable for :py:func:`scipp.transform_coords` to convert frames
    to time-of-flight.

    The graph requires the following input nodes:

    - event_time_zero and event_time_offset as read from an NXevent_data group.
    - lambda_min, used as a proxy for defining where to split and unwrap frames.
    - frame_length and frame_offset defining the frame structure.

    This assumes elastic scattering, or at least that the gaps between frames arriving
    at detectors are sufficiently large such that a common lambda_min definition is
    applicable.
    """

    def _tof_min(*, Ltotal, lambda_min):
        return _tof_from_wavelength(Ltotal=Ltotal, wavelength=lambda_min)

    def _time_offset_pivot(*, tof_min, frame_length, frame_offset):
        return _tof_to_time_offset(tof=tof_min,
                                   frame_length=frame_length,
                                   frame_offset=frame_offset)

    def _tof(*, event_time_offset, time_offset_pivot, tof_min, frame_length):
        return _time_offset_to_tof(time_offset=event_time_offset,
                                   time_offset_pivot=time_offset_pivot,
                                   tof_min=tof_min,
                                   frame_length=frame_length)

    graph = {}
    graph['tof_min'] = _tof_min
    graph['time_offset_pivot'] = _time_offset_pivot
    graph['tof'] = _tof
    return graph


def make_frames(da, *, frame_length, frame_offset=None, lambda_min=None):
    da = da.copy(deep=False)
    # TODO Should check if any of these exist, raise if they do
    da.attrs['frame_length'] = frame_length
    if frame_offset is not None:
        da.attrs['frame_offset'] = frame_offset
    if lambda_min is not None:
        da.attrs['lambda_min'] = lambda_min
    graph = Ltotal(scatter=True)
    graph.update(to_tof())
    if 'tof' in da.bins.meta or 'tof' in da.meta:
        raise ValueError("Coordinate 'tof' already defined in input data array. "
                         "Expected input with 'event_time_offset' coordinate.")
    return da.transform_coords('tof', graph=graph)
